"""
Api wrapper:
This module exports the backend APIs needed for frontend.
In the backend most of the functionality is defined within class. Frontend
should not create any object to add/modify any storage settings.
This single file should handle all of the object creation. Frontend will use
this api list to all of its access to backend. For example if frontend needs to add
a NFS share it should only call a single function add_nfs_share without creating
a NfsExports object inside front end code. This module will eventually add more contents
as backend functions grows. StorageManager is a single class for now to handle front end request
for changing/managing storage settings(raid, nfs etc).
TODO: Get more API list from front end and add more functionality here.

Usage:
from api_wrapper import StorageManager
# to check if the backend storage is available
if StorageManager.is_storage_available():
    blah
"""
__copyright__ = "Copyright 2017 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
import psutil
from backend.raid_util import RaidManager, RaidArray, RaidBackDev, RaidState
from backend.common import RAID_ARRAY_NAME
from backend.log_util import get_logger, set_log_config
from backend.lvm_util import get_volume_group_stats
LOG = get_logger('lsa.storage')


class StorageManager(object):

    @staticmethod
    def get_fake_storage_array(arrayName=RAID_ARRAY_NAME):
        ra = RaidArray('md0', 'active', 'raid10')
        ra.vg_stats = {'vg_size': 1319413953331, 'vg_free': 1002754604532, 'status': 'active'}
        ra.addDevice(RaidBackDev('/dev/sdb', 0, 'active'))
        ra.addDevice(RaidBackDev('/dev/sdc', 1, 'faulty'))
        ra.addDevice(RaidBackDev('/dev/sdd', 2, 'removed'))
        ra.addDevice(RaidBackDev('/dev/sde', 3, 'spare'))

        return ra

    @staticmethod
    def get_storage_array(arrayName=RAID_ARRAY_NAME):
        ras = RaidManager()
        ras.read_mdstat()
        return ras.get_raid_array_by_name(arrayName)

    @staticmethod
    def is_storage_available(raid=RAID_ARRAY_NAME):
        """
        check if the raid is properly setup, active and share is mounted
        and available for NFS use.
        parameter:
        raid : string
            name of the raid device
        output: True/False
        """

        ra = None
        try:
            ra = StorageManager.get_storage_array(raid)
        except Exception, e:
            LOG.exception('get storage array failed: %s' % str(e))
            return False
        if ra is None:
            LOG.error("No raid found in this system")
            return False
        if not ra.is_functional():
            LOG.error("Raid %s is not functional" % ra.dev)
            return False
        # Check volume group
        try:
            vg_stat = get_volume_group_stats()
        except Exception as e:
            LOG.warning('Failed to retrieve volume group information: %s' % str(e))
            return False
        if not vg_stat.get('vg_size') or not vg_stat.get('vg_free'):
            return False
        return True

    @staticmethod
    def get_raid_state(raid=RAID_ARRAY_NAME):
        """
        Get the raid state
        parameter:
        raid: string (optional)
        output:
            0 : unavailable
            1 : active clean
            2 : degraded
            3 : failed/faulty
            for the flag, see RaidState class
        """
        ra = None
        try:
            ra = StorageManager.get_storage_array(raid)
        except Exception, e:
            LOG.exception('get storage array failed: %s' % str(e))
        if ra in None:
            return RaidState.RAID_STATE_UNAVAILABLE
        return ra.get_raid_state()

    @staticmethod
    def get_filesystem_usage(mountpoint='/'):
        ''' returns usage information about a mounted filesystem
            when not specify return usage of root filesystem
            see psutil.disk_usage()
            named tuple including total, used and free space expressed in bytes, plus the percentage usage
        '''
        try:
            return psutil.disk_usage(mountpoint)
        except Exception:
            LOG.exception('get_filesystem_usage for %s failed' % mountpoint)
            raise


def _run_all_test():
    """
    Running some unit tests
    """
    print "Storage Availability: " + str(StorageManager.is_storage_available())
    if StorageManager.is_storage_available():
        res = StorageManager.get_filesystem_usage()
        if res:
            print "root mountpoint info: %s%%" % res.percent
        else:
            print "root mountpoint info: error while retrieving the info"
        print "/shares/ : mounted " + str(StorageManager.is_mounted("/shares/"))


def _dump_raid_array(arrayName):
    # ra = StorageManager.get_fake_storage_array()
    ra = StorageManager.get_storage_array(arrayName)
    if not ra:
        print 'no array found'
    else:
        print ra


if __name__ == '__main__':
    set_log_config()
    _run_all_test()
    _dump_raid_array(RAID_ARRAY_NAME)
