"""
Bash command:
This module executes bash/shell commands.
Usage example:
from command_util import  *
command_output_lines('rm ','-rf', '/path')
output_shell_command('cat /proc/cpuinfo | grep '' | awk -F {print $1 $2 }')
"""
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
from subprocess import Popen, PIPE
import subprocess
from backend.common import SUCCESS
from backend.log_util import get_logger, set_log_config

__all__ = ['execute_command',
           'execute_shell_command',
           'run_command_no_output',
           'command_output_lines',
           'output_shell_command'
           ]

LOG = get_logger('lsa.commands')


def get_shell_command(command):
    """
    return a shell command process for executing shell command
    parameters:
    command : string
            any bash command (ls, mkdir etc), a binary to be executed as string
    output: return what Popen returns, enabling shell

    """
    try:
        LOG.debug('executing %s' % command)
        command_process = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
    except (ValueError, OSError) as eos:
        LOG.error("Failed to create shell command process: " + str(eos))
        raise Exception("Failed to create shell command process: " + str(eos))

    return command_process


def execute_command(arguments, **options):
    """
    execute command and grab the output and return code
    param:
      arguments: command line as list of strings
      options: execution option  as dict
             expected keyword:
                  input: sub process input as string
    return: output, err and return code
    examples:
       execute_command(['foo', 'bar'])
       execute_command(['foo', 'bar'], input='hello')
    """
    if not isinstance(arguments, list) and not isinstance(arguments, tuple):
        raise TypeError('first argument must be a list')

    if not isinstance(options, dict):
        raise TypeError('second argument must be a dictionary')

    LOG.debug('executing %s' % arguments)
    if options.get('input') is not None:
        process = Popen(
            arguments,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True)
    else:
        process = Popen(
            arguments,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True)

    out, err = process.communicate(options.get('input'))
    LOG.debug('executed %s, return code %d' % (arguments, process.returncode))
    return out, err, process.returncode


def get_output_from_command(arguments):
    """
    Calling execute_command function and handling the exception
    parameter:
    arguments: list of string
            command with options ()['cat', '/path']
    output: string output of the command
    raise:
      Exception : see subprocess.CalledProcessError
    """
    out, err, ret = execute_command(arguments)
    if ret:
        cmd = arguments[0]
        error = subprocess.CalledProcessError(ret, cmd)
        error.output = out + err
        raise error
    return out


def execute_shell_command(line):
    """
    execute a shell command, ignoring output
    those command where we don't need an output, but we need the
    success/failure result
    parameters:
    line    : string
            containing full command ('mdkir /path', 'rm -rf /path' etc)
    output: Success or raise exception
    Caller is responsible for handling the erroneous result.
    """
    try:
        shell_command = get_shell_command(line)
        shell_command.wait()
    except:
        raise Exception("Shell command failed to execute: " + line)
    if shell_command.returncode != 0:
        LOG.error("Shell command failed to execute: " + line)
        return shell_command.returncode
    return SUCCESS


def run_command_no_output(args):
    """
    execute command without any output
    parameter:
    arguments : list of string
                command with options for example ['rm ','-rf', '/path' ]
    output: Success or raise exception
    Caller is responsible for handling the erroneous result.
    """
    try:
        subprocess.check_call(args)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to execute command [%s] code %d " % (e.cmd, e.returncode))
        raise Exception("Failed to execute command [%s] code %d " % (e.cmd, e.returncode))

    return SUCCESS


def command_output_lines(args):
    """
    Execute command and split the lines and return list of lines
    parameters:
    arguments : list of string
                command with options for example ['rm ','-rf', '/path' ]
    output: list of string for success, empty for no output, otherwise raise exception
    raise:
      Exception : command execution failed
    """

    out = get_output_from_command(args)

    if len(out) > 0:
        return out.split('\n')
    return ""


def output_shell_command(line):
    """
    Output the lines of a shell command
    parameter:
    line: string
        full command with all the options whatever shell supports
    output: return string output of the command for success,
            if any return code error, still we are returning the output.
            caller should handle the output
    """
    try:
        shell_command = get_shell_command(line)
        (output, err) = shell_command.communicate()
    except:
        raise StandardError("Failed to execute shell command: %s" % line)

    if shell_command.returncode != 0:
        LOG.error("Shell command returned failure: %s return code: %d" %
                  (line, shell_command.returncode))
        raise StandardError("Failed to execute shell command: %s" % line)
    return output


def run_all_test():
    """
    Not implemented yet
    """
    pass


if __name__ == '__main__':
    set_log_config()
    run_all_test()
