"""
File systems related utility functions that includes mount/umount, usage details etc.
Usage:

Statistics of a share:
from fs_util import ShareStat

stat = ShareStat.get_share_stat(name)
print(stat.get_size)

# mount logical volume( must run as root)
from fs_util import mount_lv

mount_lv(share_name)

...
...

"""
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."


import os
from backend.command_util import execute_command, command_output_lines
from backend.common import (CMD_BLKID, CMD_MKFS, LSA_DEFAULT_FS, MOUNT,
                            UMOUNT, PROC_MOUNT, DF, SUCCESS, CMD_XFS_GROWFS
                            )
from backend.log_util import get_logger, set_log_config
from backend.utility import get_full_lv_path, get_proc_mount_lv_path, get_full_mount_path

LOG = get_logger('lsa.fs_util')


def get_share_stat(name):
    '''
    Get the status (used, free , percentile used info by DF command)
      parameter:
        name: share name
        output: dictionary {'name': name, 'size': size, 'free': free, 'fs': fs, 'mounted': mounted}
    '''
    lv_path = get_full_lv_path(name)
    # Gather DF info
    lines = command_output_lines([DF, '-T', lv_path])
    if len(lines) < 2:
        raise Exception('df command output insufficient. Can not parse it')
    '''
    first line is column headers, so only care 2nd line
      Line format:
        Filesystem            Type 1K-blocks  Used Available Use% Mounted on
        /dev/mapper/lsa-test1 xfs   20961280 33824  20927456   1% /shares/test1
    '''
    LOG.debug('DF output: %s' % lines)
    parts = lines[1].split()

    try:
        free = int(parts[4]) * 1024
        fs = parts[1]
        mount_path = None
        if get_full_mount_path(name) == parts[6]:
            mount_path = parts[6]
        size = int(parts[2]) * 1024
        LOG.debug('stat for %s, size=%d free=%d, fs=%s,mnt=%s' % (name, size, free, fs, mount_path))
    except IndexError as e:
        raise Exception("Failed to retrieve output from df command, index error: " + str(e))
    return {'name': name, 'size': size, 'free': free, 'fs': fs, 'mount_path': mount_path}


def get_lv_fs(name):
    '''
    Get the formatted file system of a logical volume
    parameter:
        name: Name or the logical volume name or share name
        output: filesystem name (xfs, etx etc), we are returning none if no file
                system found on the logical volume. Typically code returns an error
                in that case.
    '''
    cmd = [CMD_BLKID, '-o' 'value', '-s', 'TYPE', get_full_lv_path(name)]
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('get_lvm_fs failure %s' % err)
        return None
    return out.replace('\n', '').replace(' ', '')


def format_lv_with_xfs(name):
    '''
    Format logical volume with xfs file system
      parameter:
          name: share/lv name
          output:  nothing on success / raise exception on failure
    '''
    cmd = [CMD_MKFS, '-f', get_full_lv_path(name)]
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('xfs format to lv (%s) failure: %s' % (name, str(err)))
        raise StandardError('xfs format to lv (%s) failure: %s' % (name, str(err)))
    # Formatted the lv, check the file system
    if get_lv_fs(name) != LSA_DEFAULT_FS:
        LOG.error("Invalid filesystem on the lv %s" % name)
        raise StandardError("Invalid filesystem on the lv %s" % name)


def mount_lv(name, inode32, nouuid=False):
    '''
    Mount a logical volume into /shares/name. The logical volume should be
    /dev/lsa/name. Here lsa is the volume group.
      parameter:
        name: share/lv name
        inode32 : true if the inode32 option need to be set to workaround
                  legacy application limitation
        nouuid: true to avoid t check for double mounted file systems using the
                file system uuid. As in the case for lvm snapshots
        output:  nothing on success / raise exception on failure
    '''
    mp = get_full_mount_path(name)
    if not os.path.exists(mp):
        raise StandardError("Mount point(%s) not exist" % mp)
    lv_path = get_full_lv_path(name)
    cmd = [MOUNT, lv_path, mp]
    if inode32:
        cmd += ['-o', 'inode32']
    if nouuid:
        cmd += ['-o', 'nouuid']
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('Failed to mount lvm (%s): %s' % (name, str(err)))
        raise StandardError('Failed to mount lvm (%s): %s' % (name, str(err)))
    return SUCCESS


def umount_lv(name):
    '''
    Unmount a logical volume. We have mount point in /shares/share_name, and the device
    name /dev/lsa/share_name. Once we delete a share we also unmount the logical volume.
      parameter:
        name: share/lv name
        output: nothing on success / raise exception on failure
    '''
    lv_path = get_full_lv_path(name)
    cmd = [UMOUNT, lv_path]
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('Failed to unmount lvm (%s): %s' % (lv_path, str(err)))
        raise StandardError('Failed to umount lvm (%s): %s' % (lv_path, str(err)))
    return SUCCESS


def is_lv_mounted(name):
    '''
    Check if the share is mounted
    parameter:
        name: share name in short format(only name not the full path)
        output: True or false
    '''
    mp = get_mount_point(get_proc_mount_lv_path(name))
    return mp == get_full_mount_path(name)


def get_mount_point(dev):
    '''
    Returns the mount point of a device. We return None if device is not mounted
    i.e no entry found in the /proc/mount file. Here we are not checking
    the mount point. Caller is responsible for that type of checking
    Parameter:
        dev: device name for example /dev/sda or /dev/sda1
        output: mount point, None when no entry is found in /proc/mount: otherwise raise exception
    '''
    try:
        fobj = open(PROC_MOUNT)
        lines = fobj.readlines()
    except (OSError, IOError) as e:
        raise Exception("Failed to open the file: %s , error %s" % (PROC_MOUNT, str(e)))
    finally:
        if fobj:
            fobj.close()
    for line in lines:
        line = line.strip()
        if not line:
            continue
        if dev in line:
            parts = line.split()
            if parts[0] == dev:
                return parts[1]


def resize_xfs(name):
    '''
    Resize the filesystem after logical volume extension
      parameter:
        name: name of the share or logical volume name, actually
                we keep the same name for nfs share and logical volume
        output: nothing on success / raise exception on failure
    '''
    cmd = [CMD_XFS_GROWFS, '-d', get_full_mount_path(name)]
    out, err, code = execute_command(cmd)
    if code != 0:
        # Log this as a warning, still share available with previous size
        LOG.warning('Adjust xfs filesystem size error: %s' % str(err))
        raise StandardError('Adjust xfs filesystem size error: %s' % str(err))


if __name__ == '__main__':
    set_log_config()
