"""
fstab manager:
This module manages entries in /etc/fstab file(add/remove entries).

Usage:
Add an entry:
##by calling static method:
FstabManager.add_fstab_entry(share_name)
Remove and entry:
FstabManager.remove_fstab_entry(share_name)
# If you want to create an object of FstabManager and want to use that
# object to add/remove fstab entry
# see add_fstab_entry remove_fstab_entry static functions
"""

__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

import os
import shutil
import json
import StringIO
from backend.log_util import get_logger, set_log_config
from backend.common import (FSTAB_FILE, BACKUP_FSTAB, TEMP_FSTAB_BACK,
                            LSA_DEFAULT_FS
                            )
from backend.utility import get_full_lv_path, get_full_mount_path
from backend.lvm_util import list_vg_disks

"""
cache the value for better performance
"""
_fstab_default_options = ''

LOG = get_logger('lsa.fstab')


class FstabManager(object):
    '''
    This class manages the fstab entry in /etc/fstab file. Functions of this class
    include read/write entries in fstab file
    '''
    def __init__(self, fstab_file=FSTAB_FILE):
        # fstab file where we keep the device and mount points and options for
        # automatic mount during system init. Most Linux based system uses /etc/fstab
        self.fstab_file = fstab_file
        # backup file where we create a copy of original fstab file before update
        self.back_fstab_file = BACKUP_FSTAB
        self.entries = []

    def read_fstab(self):
        '''
        Read the fstab file and load the list(entries) with FstabEntry or FstabComment
        '''
        del self.entries[:]
        try:
            fo = open(self.fstab_file, "r")
        except IOError as e:
            LOG.error("could not open %s: %s" % (self.fstab_file, str(e)))
            raise Exception("could not open %s: %s" % (self.fstab_file, str(e)))

        lines = None
        try:
            lines = fo.readlines()
        except IOError as e:
            LOG.error("Failed to read %s: %s" % (self.fstab_file, str(e)))
            raise Exception("Failed to read %s: %s" % (self.fstab_file, str(e)))
        finally:
            if fo:
                fo.close()
        try:
            for line in lines:
                # remove space
                line = line.strip()
                # remove new line
                line = line.strip("\n")
                # skip new lines or None
                if not line:
                    continue

                if line.startswith('#'):
                    entry = FstabComment(line)
                else:
                    entry = FstabEntry.create_object_from_line(line)
                # Here we have either an entry or comment. Append the entry to object
                self.entries.append(entry)
        except Exception as e:
            LOG.error('fstab parsing error. Invalid entry or file corrupted: %s, line: %s' % (str(e), line))
            raise StandardError('fstab parsing error. Invalid entry or file corrupted: %s line: %s' % (str(e), line))

    def get_entry_from_share_name(self, name):
        '''
        return the entry object if it exists
          parameter:
            name: name of the share
            output: Entry if it exist or None otherwise
        '''
        # Get the logical volume path from share name
        lv_path = get_full_lv_path(name)
        # Get the mount point from share name
        mp = get_full_mount_path(name)
        entry = FstabEntry(lv_path, mp)
        return self.get_entry(entry)

    def get_entry(self, entry):
        '''
        Check if an entry exists in the entries list
          parameter:
            entry: FstabEntry object
            output: full FstabEntry if it exists None otherwise
        '''
        total = len(self.entries)
        for ent in self.entries[:total]:
            if isinstance(ent, FstabComment):
                continue
            if ent == entry:
                return ent
        return None

    def entry_exists(self, entry):
        '''
        Check if an entry exists in the entries list
          parameter:
            entry: FstabEntry object
            output: True or False
        '''
        if self.get_entry(entry):
            return True
        return False

    def share_exists(self, shareName):
        '''
        Check if a an entry with share name exists in fstab file
          parameter:
            share: share name
            output: True or False
        '''
        # Get the logical volume path from share name
        lv_path = get_full_lv_path(shareName)
        # Get the mount point from share name
        mp = get_full_mount_path(shareName)
        entry = FstabEntry(lv_path, mp)
        return self.entry_exists(entry)

    def is_inode32_set(self, share_name):
        '''
        Check if inode32 is set for a share name in fstab file
          parameter:
            share: share name
            output: True or False
        '''
        entry = self.get_entry_from_share_name(share_name)
        if entry:
            options = entry.get_options()
            if options and 'inode32' in options:
                return True
        return False

    def add_share(self, name):
        '''
        Add a share to fstab
        '''
        entry = FstabEntry.create_from_share_name(name)
        self.add_entry(entry)

    def remove_share(self, name):
        '''
        Remove a share from fstab
        '''
        entry = FstabEntry.create_from_share_name(name)
        self.remove_entry(entry)

    def get_all_entries_as_str(self):
        '''
        Get the entries as string. Each entry in a separate line
        '''
        output = StringIO.StringIO()
        for entry in self.entries:
            output.write(str(entry) + '\n')
        contents = output.getvalue()
        output.close()
        return contents

    def add_entry(self, entry):
        '''
        Add an entry to fstab file
          parameter:
              entry: FstabEntry/FstabComment object
              output: SUCCESS or raise exceptions
        '''
        self.read_fstab()
        return self._append_entry(entry)

    def remove_entry(self, entry):
        '''
        Remove an entry from fstab file
          parameter:
              entry: FstabEntry/FstabComment object
              output: nothing on success or raise exceptions for failure
        '''
        found = False
        self.read_fstab()
        for idx, ent in enumerate(self.entries):
            # We only check actual entry. skip comments
            if isinstance(ent, FstabComment):
                continue
            if ent == entry:
                found = True
                break
        if not found:
            return
        try:
            del self.entries[idx]
            self._write_all_entries()
        except Exception as e:
            LOG.error('Failed to remove an entry(%s) from fstab: %s' % (str(entry), str(e)))
            raise StandardError('Failed to remove an entry(%s) from fstab: %s' % (str(entry), str(e)))

    def _write_all_entries(self):
        '''
        write list of entries to fstab file. In order to do an atomic update
        we first write to a temporary file and then rename/replace the original fstab file
        with temporary file. Before replacing the original file we also create a backup
        of the original fstab file.
        '''
        LOG.info("Writing fstab entries")
        temp_file = TEMP_FSTAB_BACK
        if os.path.exists(temp_file):
            LOG.warning('Unusual case, temp file(%s) exists.Removing temp file' % temp_file)
            os.remove(temp_file)
        # create new file
        bfo = None
        try:
            bfo = open(temp_file, "w")
        except IOError as e:
            LOG.error("Failed to create the backup file:" + temp_file)
            raise Exception("Failed to create the backup file: %s, %s" % (temp_file, str(e)))
        try:
            bfo.write(self.get_all_entries_as_str())
        except IOError as e:
            LOG.error("Failed to write entries to file : \n" + str(e))
            raise Exception("Failed to write fstab entries: " + str(e))
        finally:
            if bfo:
                bfo.close()
        LOG.info("exports entries written to %s" % temp_file)

        # create a backup file just in-case. at first remove the backup file
        if os.path.exists(self.back_fstab_file):
            os.remove(self.back_fstab_file)
        # Copy the original
        shutil.copyfile(self.fstab_file, self.back_fstab_file)
        # Finally move the temp fstab file to original
        shutil.move(temp_file, self.fstab_file)

    def _append_entry(self, entry):
        """
        append an entry fo /etc/fstab file.
          parameter:
              entry: FstabEntry or Fstab comment object
              output: Success or raise exception
        """

        if self.entry_exists(entry):
            LOG.error("Trying to add duplicate entry")
            raise Exception("Trying to add duplicate fstab entry: " + str(entry))
        bfo = None
        try:
            bfo = open(self.fstab_file, "a+")
        except IOError:
            LOG.error("Failed to open the file for editing: %s", self.fstab_file)
            raise Exception("Failed to open the file for editing: %s", self.fstab_file)
        # create a backup file just in-case
        if os.path.exists(self.back_fstab_file):
            os.remove(self.back_fstab_file)
        shutil.copyfile(self.fstab_file, self.back_fstab_file)
        try:
            bfo.write(str(entry))
            bfo.write("\n")
        except IOError as e:
            LOG.error("Failed to append entry: \n" + str(e))
            raise Exception("Failed to append entry: %s" % str(e))
        finally:
            if bfo:
                bfo.close()
        # Finally append the entry to the list
        self.entries.append(entry)

    @classmethod
    def get_inode32_for_share(cls, name):
        '''
        This is a static method to retrieve inode32 settings for a share
        parameter:
            name: name of the share
        output: True if inode32 is set, False otherwise
        '''
        fstab_mgr = cls(FSTAB_FILE)
        fstab_mgr.read_fstab()
        return fstab_mgr.is_inode32_set(name)

    @classmethod
    def share_exist_in_fstab(cls, name):
        '''
        This is a static method to retrieve inode32 settings for a share
        parameter:
            name: name of the share
        output: True if inode32 is set, False otherwise
        '''
        fstab_mgr = cls(FSTAB_FILE)
        fstab_mgr.read_fstab()
        return fstab_mgr.share_exists(name)

    @classmethod
    def add_fstab_entry(cls, name, inode32):
        '''
        This is a static method to add an entry to fstab file
        parameter:
            name: name of the share
            output: nothing on success or raise exception on failure
            inode32 : true if the inode32 option need to be set to workaround
                      legacy application limitation
        '''
        # Get the logical volume path from share name
        lv_path = get_full_lv_path(name)
        # Get the mount point from share name
        mp = get_full_mount_path(name)
        fstab_mgr = cls(FSTAB_FILE)
        options = None
        if inode32:
            global _fstab_default_options
            if _fstab_default_options == '':
                cls.set_fstab_default_options()
            options = _fstab_default_options + ',inode32'
        try:
            entry = FstabEntry(lv_path, mp, options=options)
            fstab_mgr.add_entry(entry)
        except Exception as e:
            LOG.error('Failed to add an entry(%s) to fstab file: %s' % (name, str(e)))
            raise StandardError('Failed to add an entry(%s) to fstab file: %s' % (name, str(e)))

    @classmethod
    def remove_fstab_entry(cls, name):
        '''
        This is a static method to remove an entry from fstab file
        parameter:
            name: name of the share
            output: Success or raise exception
        '''
        # Get the logical volume path from share name
        lv_path = get_full_lv_path(name)
        # Get the mount point from share name
        mp = get_full_mount_path(name)
        fstab_mgr = cls(FSTAB_FILE)
        try:
            entry = FstabEntry(lv_path, mp)
            fstab_mgr.remove_entry(entry)
        except Exception as e:
            LOG.error('Failed to remove an entry(%s) to fstab file: %s' % (name, str(e)))
            raise StandardError('Failed to remove an entry(%s) to fstab file: %s' % (name, str(e)))

    @classmethod
    def set_fstab_default_options(cls):
        '''
        This is a static method to set the global variable _fstab_default_options
        _netdev and nofail are needed for iSCSI remote block volumes, as recommended at
        https://docs.us-phoenix-1.oraclecloud.com/Content/Block/References/fstaboptions.htm
        '''
        global _fstab_default_options
        _fstab_default_options = 'defaults'
        try:
            vg_disks = list_vg_disks()
        except Exception as e:
            LOG.error('Failed to retrieve volume group information: %s' % str(e))
            raise StandardError('Failed to retrieve volume group information: %s' % str(e))
        if not (vg_disks[0].startswith('/dev/md') or vg_disks[0].startswith('/dev/nvme')):
            _fstab_default_options += ',_netdev,nofail'


class FstabComment(object):
    '''
    An entry in fstab file which represents a comment.
    We are using generic fstab file which is also used by other system modules.
    Keeping the comments make sense here.
    '''
    def __init__(self, comment):
        '''
        Comment must start with '#'
        '''
        if not comment.startswith('#'):
            raise ValueError('Comment must start with #')
        self.comment = comment

    def __str__(self):
        return self.comment

    def __eq__(self, other):
        return self.comment == other.comment


class FstabEntry(object):
    '''
    An entry in fstab file
    dev: device path
    mp: mount point
    fs: file system
    options: comma separated options
    dp: option that need to be used by the dump program. If the value is set to
        0, then the partition is excluded from taking backup and if
        the option is a nonzero value, the device will be backed up.
    fsck:mentions the fsck option. That is if the value is set to zero,
        the device or partition will be excluded from fsck check and if it is
        nonzero the fsck check will be run in the order in which the value
        is set. The root partition will have this value set to one so that it
        will be checked first by fsck.
    '''
    def __init__(self, dev, mp, fs=LSA_DEFAULT_FS, options=None, dp=0, fsck=0):
        self.mp = mp
        self.fs = fs
        if not options:
            global _fstab_default_options
            if _fstab_default_options == '':
                FstabManager.set_fstab_default_options()
            options = _fstab_default_options
        self.options = options
        self.dp = dp
        self.fsck = fsck
        self.dev = dev

    def __eq__(self, other):
        return self.dev == other.dev

    def __str__(self):
        return "{} {} {} {} {} {}".format(self.dev, self.mp, self.fs, self.options, self.dp, self.fsck)

    def get_dev(self):
        '''
        Get the device name of a fstab entry
        '''
        return self.dev

    def get_mp(self):
        '''
        Get mount point of the fstab entry
        '''
        return self.mp

    def get_fs(self):
        '''
        Get file system name of a fstab entry
        '''
        return self.fs

    def get_options(self):
        '''
        Get mount options of a fstab entry
        '''
        return self.options

    def get_dp(self):
        '''
        Get dump program option
        '''
        return self.dp

    def get_fsck(self):
        '''
        Get FSCK option
        '''
        return self.fsck

    @classmethod
    def create_from_share_name(cls, name):
        '''
        Create an object from a share name
          parameter:
            name: name of the share
            output: FstabEntry type object
        '''
        # Get the logical volume path from share name
        lv_path = get_full_lv_path(name)
        # Get the mount point from share name
        mp = get_full_mount_path(name)
        return FstabEntry(lv_path, mp)

    @classmethod
    def create_object_from_line(cls, line):
        '''
        Create an object from a line in specific format:
          parameter:
            line: line as string
                  format: /dev/test1 /shares/test1 xfs default 0 0
            output: FstabEntry type object
        '''
        # We are using * to get the value from filter as filter(lambda x: x not in ('', None), line.split(" "))
        # returns reference
        return cls(*filter(lambda x: x not in ('', None), line.split(" ")))

    @classmethod
    def create_object_from_json(cls, js):
        """
        create an object from json format data
        parameter:
          input:
            js: json format
                {'dev': '/shares/jsontest3',
                 'mp': 'jsontest3',
                 'fs': 'xfs',
                 'options': '',
                 'dp': '0',
                 'fsck': '0'
                }
          output: FstabEntry object
         """
        deserializeJS = json.loads(js)
        # We only care about device and mount point. Other parameters are default
        # we can safe ignore those
        fe = cls(deserializeJS.get('dev'), deserializeJS.get('mp'))
        return fe


if __name__ == '__main__':
    set_log_config()
    print FstabManager.get_inode32_for_share('node2')
