#!/usr/bin/bash
# Copyright (c) 2017, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.

: '
    This script configures samba for lsa server
     Copy default smb config file

    This script is called from main.sh
    If you want to run independently  use ./samba.sh
'

CUR_DIR="$(dirname "$0")"
source $CUR_DIR/globals.sh
source $CUR_DIR/common.sh

: '
    Setup selinux rules for samba generic access i.e for smbd service to access shares
    in the system. Helpful link: https://selinuxproject.org/page/SambaRecipes
'
function setup_samba_selinux_rules() {


    $SETSEBOOL -P samba_export_all_ro=1
    check_exit "smb setsebool samba_export_all_ro"
    out=$((out+$?))
    $SETSEBOOL -P samba_export_all_rw=1
    check_exit "smb setsebool samba_export_all_rw"
    out=$((out+$?))
    $SETSEBOOL -P smbd_anon_write=1
    check_exit "smb setsebool smbd_anon_write"
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1
};

: '
  This function takes care of SELinux rules for samba export/resource file
'
function setup_samba_pool_se_config() {
    out=0

    $SEMANAGE fcontext -a -t public_content_t ${SMB_EXPORT_CONF}
    check_exit "semanage lsa.conf"
    out=$((out+$?))
    $RESTORECON -F -R -v ${SMB_EXPORT_CONF}
    check_exit "restorecon lsa.conf"
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1
};

: '
  Apply SElinux rules to export shares
  Add read/write access SELinux rules for samba export of shares
'
function setup_samba_shares_se_config() {
    echo "Locating all CIFS resources in ${SMB_EXPORT_CONF}"
    # read smb export conf file and filter on 'path=' 
    # apply SELinux rules to all resources path
    # resource can be duplicated: path (shares) can be exported multiple time 
    ${AWK} -F' = ' ' $1 == "path" {all_paths[$2]=0} END {for (path in all_paths) print path}' < ${SMB_EXPORT_CONF} | while read share_path
    do
	echo "Applying SELinux rules to ${share_path}"
	${CHCON} --type=samba_share_t ${share_path}
	if [ $? -ne 0 ]
	then
	    echo "WARNING: cannot change context on ${share_path}"
	fi
	${SEMANAGE} fcontext --add --type samba_share_t "${share_path}(/.*)?"
	if [ $? -ne 0 ]
	then
	    echo "WARNING: cannot manage policy on ${share_path}"
	fi
	${SEMANAGE} fcontext --add --type public_content_rw_t "${share_path}(/.*)?"
	if [ $? -ne 0 ]
	then
	    echo "WARNING: cannot manage policy on ${share_path}"
	fi
	${RESTORECON} -R -v ${share_path}
	if [ $? -ne 0 ]
	then
	    echo "WARNING: cannot restore context on ${share_path}"
	fi
    done
    return 0
}


function setup_samba_pool_config() {
    if [ ! -d ${CONF_MNT}/etc/smb ]
    then
	${MKDIR} --parents ${CONF_MNT}/etc/smb
	if [ $? -ne 0 ]
	then
	    echo "Cannot create directory ${CONF_MNT}/etc/smb" >&2
	    return 1
	fi
    fi
    if [ ! -f $SMB_EXPORT_CONF ]
    then
	echo "creating share SMB conf file"
	$TOUCH $SMB_EXPORT_CONF
	if [ $? -ne 0 ]
	then
	    echo "Cannot create file $SMB_EXPORT_CONF" >&2
	    return 1
	fi
    fi
    return 0
}

: '
    Copy default samba config file from lsa source to /etc/samba
'
function copy_default_config() {


    out=0
    echo "Make the copy of original samba config file"
    $CP -f $SMB_CONF ${SMB_CONF}.orig
    check_exit "copy original smb conf"
    out=$((out+$?))
    echo "Copying default config file from lsa source"
    $CP -f $DEFAULT_SRC_SMB_CONF $SMB_CONF
    check_exit "copy smb from lsa"
    out=$((out+$?))

    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1
};

function main_setup_samba() {

    echo "Configuring samba for LSA"
    copy_default_config
    out=$((out+$?))
    setup_samba_selinux_rules
    out=$((out+$?))
    if [ "$out" == "0" ]; then
       return 0
    fi
    return 1
};

script_name=$($BASENAME ${0#-}) #- needed if sourced no path
this_script=$($BASENAME ${BASH_SOURCE})

# If running this script should call the main function,
# Other wise script is used as a source
if [[ ${script_name} = ${this_script} ]] ; then
    out=0
    main_setup_samba
    out=$((out+$?))
    setup_samba_pool_config
    out=$((out+$?))
    setup_samba_pool_se_config
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	    echo "  Samba setup successful"
	    exit 0
    fi
    echo "Samba setup failed"
    exit 1
fi
