"""
Manage types commonly used by other modules
"""
__copyright__ = "Copyright 2017 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

import Queue
import cStringIO
import json

__all__ = ['StreamQueue', 'StreamPart', 'EOSPart', 'ExtJSONEncoder']


class StreamQueue(Queue.Queue):
        """
        Sub class of Queue.Queue
        Brings capabilities to fetch status of the queue
        It allows the producer to warn consumers about the fact
        that no more element will be pushed
        """
        def __init__(self, max):
                Queue.Queue.__init__(self, max)
                self.active = True

        def is_active(self):
                """
                Returns true if this instance is active.
                I.e consumers can expect new element to be posted
                """
                return self.active

        def set_active(self, is_active):
                """
                Sets the status of the queue
                parameters:
                   is_active: True or False
                """
                self.active = is_active


class StreamPart(object):
        """
        Fragment post to a StreamQueue
        """
        def __init__(self, part_num, size, buffer):
                """
                Creates a new StreamPart
                parameters:
                   part_num: part number
                   size : part size (size of the buffer in bytes)
                   buffer: instance of a stream (cStringIO)
                """
                self.part_num = part_num
                self.buffer = buffer
                self.buffer_size = size

        def __str__(self):
            return 'StreamPart{part_num=%d, size=%d}' % (self.part_num, self.buffer_size)

        def get_part_num(self):
                """
                Returns the part number as int
                """
                return self.part_num

        def get_buffer(self):
                """
                Returns the wrapped stream
                """
                return self.buffer

        def get_size(self):
                """
                Returns the size of the part
                """
                return self.buffer_size

        def cleanup(self):
                """
                cleanups
                I.e free resrouce of underlying buffer
                """
                self.buffer.close()


class EOSPart(StreamPart):
        """
        Special instance used to send a EndOfStream message to the queue
        """
        def __init__(self):
                StreamPart.__init__(self, 0, 0, None)


class ErrorPart(StreamPart):
        """
        Special instance used to send a error message to the queue
        """
        def __init__(self, error_message):
                StreamPart.__init__(self, -1, len(error_message),
                                    cStringIO.StringIO(error_message))


class ExtJSONEncoder(json.JSONEncoder):
        """
        JSON specialized encoder use to jsonify objects
        If object have a method called 'serialize'
        It will be used for serialization
        see json.dump documentation
        """
        def default(self, obj):
                if hasattr(obj, 'serialize'):
                        return obj.serialize()
                else:
                        return json.JSONEncoder.default(self, obj)
