"""
LVM manager:
This module manages logical volume(add, remove logical volumes, retrieve
volume group/logical volume information).

Usage:
********* This function must be called by program executed as root *********
Add a logical volume:
from lvm_util import add_lv

add_lv(lv_name, size)

Remove a logical volume
from lvm_util import remove_lv
remove_lv(name)

...
...

"""
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."


from backend.command_util import execute_command, command_output_lines
from backend.common import (VG_NAME, LVCREATE, LVREMOVE, LVEXTEND, PVSCAN,
                            CMD_DD, LVS, VGS, CMD_VGDISPLAY,
                            LSA_CMD_CREATE_POOL
                            )
from backend.log_util import get_logger, set_log_config
from backend.utility import get_full_lv_path, get_full_mount_path, round_to_multiplier
from backend.fs_util import get_share_stat, mount_lv, umount_lv
import os

LOG = get_logger('lsa.lvm')


def get_lv_stats(name):
    '''
    Gather volume info, lv name, VG name, vg size etc
      parameter:
        name: share/lv name
        output: logical volume stats in dictionary format {'vg_size': xxx, 'vg_free':xxx
        'name': share_name, 'size': xxx, 'free': xxx, 'fs': xfs, 'mounted': True/False}
        all units in bytes.
        raise exception on failure. Caller should handle the exception
    '''

    lv_path = get_full_lv_path(name)
    cmd = [LVS, lv_path, '--noheadings', '--nosuffix', '--units', 'b', '-o', 'name,path,lv_size,vg_name,vg_free,vg_size']
    try:
        lines = command_output_lines(cmd)
    except Exception as e:
        raise StandardError('Failed to retrieve logical volume information by lvs command: %s' % str(e))
        LOG.error('Failed to retrieve logical volume information by lvs command: %s' % str(e))
    if len(lines) <= 1:
        # Error
        raise Exception('lvs output insufficient:\n %s', lines)
    lvstat = {}
    # Format of the line: lv_name /dev/lsa/lv_name 1073741824 /dev/md0(0) lsa 126596677632 128744161280
    # {lv_name} {lv_path} {lv_size} {volume gorup name}
    # {free space of volume group in bytes} {size of the volume group}
    parts = lines[0].split()
    try:
        # parse lvs output
        lvstat['vg_size'] = int(parts[5])
        lvstat['vg_free'] = int(parts[4])
        lvstat['lv_size'] = int(parts[2])
    except IndexError as e:
        raise Exception("Failed to retrieve output from lvs command, index error: " + str(e))
    try:
        # Merging the stats got from lvs command
        lvstat.update(get_share_stat(name))
    except Exception as e:
        raise StandardError('Failed to get share status(LVS command): %s' % str(e))
    return lvstat


def add_lv(name, size):
    '''
    Add a logical volume to lsa volume group
    parameter:
    name: share name
    size: size of the share (here size is in bytes)
    output: nothing on success or raise exception on failure
    '''
    cmd = [LVCREATE, '-y', '--name', name, '--size', str(size) + 'b', VG_NAME]

    out, err, code = execute_command(cmd)

    LOG.debug('add_lv (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('add_lv failure %s' % err)
        raise StandardError('cannot add logical volume(%s): %s' % (name, err))

    LOG.debug('adding logical volume(%s) done' % name)


def add_lv_snapshot(lv_name, snapshot_name, size):
    '''
    Add a logical volume snapshot named snapshot_name referencing lv_name
    parameter:
    lv_name: share name
    snapshot_name: name of the snapshot to be created
    size: size of the share (here size is in bytes)
    output: nothing on success or raise exception on failure
    '''
    cmd = [LVCREATE, '--size', str(size) + 'b', '--snapshot', '--name', snapshot_name, get_full_lv_path(lv_name)]

    out, err, code = execute_command(cmd)

    LOG.debug('add_lv_snapshot (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('add_lv_snapshot failure %s' % err)
        raise StandardError('cannot add logical volume snapshot (%s): %s' % (snapshot_name, err))

    LOG.debug('adding logical volume snapshot (%s) done' % snapshot_name)


def remove_lv(name):
    '''
    Remove a logical volume form lsa volume group
    parameter:
        name: share name
        output: Nothing on success or raise exception on failure
    '''
    cmd = [LVREMOVE, '-f', get_full_lv_path(name)]

    out, err, code = execute_command(cmd)

    LOG.debug('remove_lv (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('remove_lv failure %s' % err)
        raise StandardError('cannot remove a logical volume(%s): %s' % (name, err))

    LOG.debug('removing logical volume(%s) done' % name)


def remove_lv_fs_signature(name):
    '''
    Wipe out file system signature from logical volume.
    Because removing the logical volume does not wipe out the FS superblock.
    So next time when we create an LV starting from the sector where previous
    FS signature exits, lvcreate raises error unless we provide -y option.
    parameter:
        name: share name
        output: Success or raise exception
    '''
    dd_of = 'of=' + get_full_lv_path(name)
    cmd = [CMD_DD, 'if=/dev/zero', dd_of, 'count=1', 'bs=1k']
    out, err, code = execute_command(cmd)

    LOG.debug('remove_fs_signature (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('remove_fs_signature failure %s' % err)
        raise StandardError('cannot wipe out FS signature on lv(%s): %s' % (name, err))

    LOG.debug('removing FS signature volume(%s) done' % name)


def extend_lv(name, new_size):
    '''
    Extend a logical volume. Here we also resize the file system, so file system
    on top of LV is required here
    parameter:
        name: name of the share
        size: new size of the share( size is in bytes)
        output: Success or raise exception
    '''
    cmd = [LVEXTEND, '-v', '--size', str(new_size) + 'b', get_full_lv_path(name)]

    out, err, code = execute_command(cmd)

    LOG.debug('extend_lv (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('extend_lv failure %s' % err)
        raise StandardError('cannot extend a logical volume(%s): %s' % (name, str(err)))

    LOG.debug('extending logical volume(%s) done' % name)


def get_volume_group_stats():
    '''
    Get size, available space of the volume group(lsa) in bytes
    Output: dictionary, format {'vg_size': 'xxxxxx', 'vg_free': 'xxxxxx', 'status': 'active/faulty'}
    raise: standard error in case of failure
    '''
    # Here vgs output is in bytes, for better accuracy of
    cmd = [VGS, VG_NAME, '--noheadings', '--nosuffix', '--units', 'b', '-o', 'vg_size,vg_free']
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('Failed to get size of vg (%s): %s' % (VG_NAME, str(err)))
        raise StandardError('Failed to get size of vg (%s): %s' % (VG_NAME, str(err)))
    # format of the output: 126596677632 128744161280: {size of volume group} {free space of volume group}
    # All the units are in bytes
    # default status is active, if we get any error from the output we set the status as faulty
    # but we still get the size and free information from the command
    status = 'active'
    if len(err) > 0:
        status = 'faulty'
    parts = out.split()
    try:
        size = parts[0]
        free = parts[1]
    except IndexError as e:
        LOG.error('Error parsing vgs output: %s' % str(e))
        raise StandardError('Error parsing vgs output: %s' % str(e))
    return {'vg_size': size, 'vg_free': free, 'status': status}


def setup_snapshot(share_to_snapshot):
    '''
    Create a lv snapshot for share_to_snapshot and mount it
    parameter:
        share_to_snapshot: str, relative name of the share to snapshot
    '''
    # Estimate snapshot lv size by calculating the maximum amount of change
    # that could be written during backup.  Assuming the streaming backup
    # throughput is 110M per second and writing to share via NFS can be done at
    # a throughput of 13M per second. One can calculate the ratio needed
    # to determine how much space should be reserved in our case 110/13
    # Assuming the worse case is unlikely, the ratio is reduced by half
    # giving ((110/13)*2)

    share_size = get_lv_stats(share_to_snapshot)['size']
    share_size = share_size / ((110 / 13) * 2)
    # set size to the closest multiple of 512, it's a lvcreate requirement
    share_size = round_to_multiplier(share_size, 512)

    share_snapshot = share_to_snapshot + '_snapshot'
    # create snapshot and mount it
    try:
        add_lv_snapshot(share_to_snapshot, share_snapshot, share_size)
        os.makedirs(get_full_mount_path(share_snapshot))
        mount_lv(share_snapshot, False, nouuid=True)
    except Exception, e:
        LOG.exception('Error cannot create snapshot: %s' % str(e))
        raise StandardError('Error cannot create snapshot: %s' % str(e))
    return share_snapshot


def remove_snapshot(snapshot):
    '''
    umount, remove lv and directory associated with snapshot
    parameter:
        snapshot: str, relative name of the snapshot to remove
    '''
    try:
        umount_lv(snapshot)
        remove_lv(snapshot)
        os.rmdir(get_full_mount_path(snapshot))
    except Exception, e:
        LOG.exception('Error cannot remove snapshot: %s' % str(e))
        raise StandardError('Error cannot remove snapshot: %s' % str(e))


def list_lvs_by_vg(vg_name=VG_NAME):
    '''
    List the logical volumes under lsa volume group (lsa)
    output: lines (each line represents a logical volume. Line format( lv_name, lv_size(bytes))
    raises: standard error
    '''
    # Here vgs output is in bytes, for better accuracy of
    # output format
    '''
    test1  3221225472
    test2 21474836480
    '''

    cmd = [VGS, vg_name, '--noheadings', '--nosuffix', '--units', 'b', '-o', 'lv_name,lv_size']
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.warning('Failed to get size of vg (%s): %s' % (VG_NAME, str(err)))
        raise StandardError('Failed to get size of vg (%s): %s' % (VG_NAME, str(err)))
    LOG.debug('vgs command on %s: %s' % (vg_name, out))
    return out.strip('\n')


def list_vg_disks():
    '''
    List the physical volumes of the volume group named lsa.
    Parameter:
        input: None
        output: list of disk format: ['/dev/sdb', '/dev/sdc', '/dev/sdd', '/dev/sde']
                Empty list if the volume group is not present
    raise: Standard error if command fails
    '''
    # List all physical volumes of the volume group lsa
    arg = [VGS, VG_NAME, '--noheadings', '-o', 'pv_name', '--rows']
    out, err, code = execute_command(arg)
    if code != 0:
        LOG.warning('Failed to list the physical volumes of volume group (%s): %s' % (VG_NAME, str(err)))
        raise StandardError('Failed to list the physical volumes of volume group (%s): %s' % (VG_NAME, str(err)))
    out = out.strip()
    out = out.strip('\n')
    return out.split()


def is_vg_created():
    '''
    Check if the volume group lsa exists.
    parameter:
        input: None
        output: boolean, True if exists otherwise False
    raise: StandardError if command fails
    '''
    arg = [CMD_VGDISPLAY, '-C', '--no-headings']
    out, err, code = execute_command(arg)
    if code != 0:
        LOG.warning('Failed to get volume group information: %s' % str(err))
        raise StandardError('Failed to get volume group information: %s' % str(err))
    if not out:
        return False

    out = out.strip('\n')
    lines = out.split('\n')
    for line in lines:
        line = line.strip()
        parts = line.split()
        # check first column is VG name i.e lsa
        if parts[0] == VG_NAME:
            return True
    return False


def get_physical_volumes():
    """
    Gets all physical volumes on the system
    Returns a list of physical volumes (device name). Can be empty
    """
    lines = None
    try:
        lines = command_output_lines([PVSCAN, '--quiet', '--uuid'])
    except Exception as e:
        raise StandardError('Failed to retrieve physical volume information: %s' % str(e))
        LOG.error('Failed to retrieve physical volume information: %s' % str(e))

    # we rely on pvscan(8) output being like
    #  PV /dev/sdb with UUID XGBCtV-auKA-LKdd-L5up-UHaf-iD0A-dYf4ve    VG lsa             lvm2 [<1024.00 GiB / <682.00 GiB free]
    #  PV /dev/sdc with UUID dbo1D5-47un-Z0ej-RH2k-PP0H-oabB-ey3W5Q                       lvm2 [50.00 GiB]
    #  PV /dev/sdd with UUID evpjaz-Sk69-DyNh-UMBY-IHJj-Icbb-taUoa1                       lvm2 [50.00 GiB]
    #  Total: 3 [<1.10 TiB] / in use: 1 [<1024.00 GiB] / in no VG: 2 [100.00 GiB]
    #
    #  We check for lines starting with 'PV' and get the device name from it
    _all_devices = []
    for line in lines:
        if len(line) == 0:
            continue
        _tokens = line.split()
        if _tokens[0] == "PV":
            _all_devices.append(_tokens[1])
    return _all_devices


def create_pool(raid_level, devices):
    """
    Create the pool
    parameters:
        raid_level: raid level (0, 1, 6, 10, or RAID_LEVEL_NONE)
        devices: list of comma-separated paths (e.g., '/dev/sdb,/dev/sdc')
    output: nothing on success or raise exception on failure
    """
    devs = [d.strip() for d in devices.split(',')]
    cmd = [LSA_CMD_CREATE_POOL, str(raid_level)]
    cmd.extend(devs)
    out, err, code = execute_command(cmd)
    LOG.debug('create_pool (%d,%s,%s)' % (code, out, str(err)))

    if code != 0:
        LOG.warning('create_pool failure: %s' % err)
        raise StandardError('cannot create pool: %s' % err)


if __name__ == '__main__':
    set_log_config()
