'''
Custom class for option type variable.

Usage:
from option_util import BooleanOption, BooleanOptionWithPrefix
read_only = BooleanOption('read_only')
read_only.disable()
print str(read_only.enabled())
'''
___copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from log_util import get_logger, set_log_config

LOG = get_logger('lsa.option_util')


class BooleanOption(object):
    '''
    This class represents an option with boolean value.
    '''
    def __init__(self, name, value=False):
        '''
        Constructor for BooleanOption
        name: option name , type: string
        val: value of the option  as boolean (default=False)
        '''
        self.name = name
        if not isinstance(value, bool):
            raise ValueError('value must be boolean')
        self.value = value

    def get_name(self):
        '''
        Return the name of the option as string
        '''
        return self.name

    def enabled(self):
        '''
        Check if the option is enabled. Return True/False
        '''
        return self.value

    def enable(self):
        '''
        Enable the option i.e set the value to True
        '''
        self.value = True
        return self

    def disable(self):
        '''
        Disable the option i.e set the value to false
        '''
        self.value = False
        return self

    def __repr__(self):
        '''
        String representation of the option, usually the name.
        If some option does not represent anything in case of disabled we
        don't return anything. For example if sync is false and we don't want to
        return the name for nfs option. Because if the value is false we don't want to include it in the
        configuration file. If you want to use prefix in case of disabled please use the BooleanOptionWithPrefix
        class.
        '''
        if self.enabled():
            return self.name
        return ''

    def to_dict(self):
        '''
        Convert this item to dictionary, this function is used if caller wants
        to use the tuple in a list or separately.
        '''
        return dict([(self.get_name(), self.value)])

    def __str__(self):
        '''
        string representation of the option
        '''
        return self.name + 'is ' + str(self.enabled())


class BooleanOptionWithPrefix(BooleanOption):
    '''
    This is the extended class of BooleanOption.
    If the value is false we add the prefix (no_/no/in) before the name
    and represent it. For example root_squash, subtree_check >> no_root_squash, no_subtree_check
    secure >> insecure
    '''
    def __init__(self, name, prefix='no_', value=False):
        '''
        Constructor for BooleanOptionWithPrefix
        name: option name
        prefix: prefix that is added to the option name if the option is disabled
        value: value of the option as boolean (default=False)
        '''
        BooleanOption.__init__(self, name, value)
        self.prefix = prefix

    def __repr__(self):
        '''
        String representation of the option, usually the name, if the option is disabled
        returns prefix + name
        '''
        name = self.get_name()
        if not self.enabled():
            name = self.prefix + name
        # 'hide' does not show up in the .exports file
        if name == 'hide':
            name = ''
        return name


if __name__ == '__main__':
    set_log_config()
