"""
samba configuration utility module:
Usage:
import backend.samba_util
# get samba options
backend.samba_util.get_samba_options()
# set samba options
backend.samba_util.set_samba_options(options)
"""

__copyright__ = "Copyright 2017, 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."


import StringIO
import json
from log_util import get_logger, set_log_config
from common import USER_NOBODY, SAMBA_DIR_ACCESS_MASK
from export_base import ExportBase
from option_util import BooleanOption
from utility import parse_boolean

LOG = get_logger('lsa.samba_util')


class SambaConfKey(object):
    '''
    This class represents the CONSTANTS for keys in the samba configuration file.
    See SambaExportEntry class for details of each key.
    More reference: https://www.samba.org/samba/docs/using_samba/ch06.html
    '''
    PATH = 'path'
    RESOURCE = 'resource'
    READ_ONLY = 'read only'
    READ_ONLY_VAR = 'read_only'
    COMMENT = 'comment'
    STRICT_SYNC = 'strict sync'
    STRICT_SYNC_VAR = 'strict_sync'
    SYNC_ALWAYS = 'sync always'
    SYNC_ALWAYS_VAR = 'sync_always'
    GUEST_OK = 'guest ok'
    GUEST_OK_VAR = 'guest_ok'
    BROWSABLE = 'browsable'
    HOSTS_ALLOW = 'hosts allow'
    HOSTS_DENY = 'hosts deny'
    HOSTS_ALLOW_VAR = 'hosts_allow'
    HOSTS_DENY_VAR = 'hosts_deny'
    DIRECTORY_MASK = 'directory mask'
    CREATE_MASK = 'create mask'
    FORCE_USER = 'force user'


class SambaExportEntry(ExportBase):

    def __init__(self, resource, **options):
        '''
        resource: resource name of the samba export. Resource is the unique property of an export
        Path might in option or not. We don't keep it within the object.
        When we get the options from UI there is no path attribute, when we get options from samba config file
        path is present but we don't use it as UI does not need it because client only care about resource to
        access/mount the samba share.
        options: dictionary items
            comment: comment or description of the share so that client can understand the type/use of the share
            read only: whether the share is read only or not (read_only also supported)
            browsable: True/False whether share is browsable in the client side either network or other source of browser
            'strict sync' | strict_sync: True/False, Samba honors the FILE_FLAG_WRITE_THROUGH flag, and results in synchronous writes when called by the CreateFile API
            'sync always' | sync_always: True/False Samba executes all writes synchronously
            'guest ok' | guest_ok: whether share is public or not. If the share is public then {hosts allow} must be empty
            'hosts allow'|hosts_allow: comma separated IPs or hostnames who are allowed to access the share
            'hosts deny'|hosts_deny: comma separated IPs or hostnames who are denied to access the share
        '''
        if resource is None:
            raise ValueError('Resource can not be None')
        if options is None:
            raise ValueError('Samba option can not be None')
        # Base class has path attribute, for samba we are storing resource name
        ExportBase.__init__(self, resource, 'smb', **options)
        if SambaConfKey.READ_ONLY in options:
            if parse_boolean(options.get(SambaConfKey.READ_ONLY)) is False:
                self.read_only.disable()
        elif SambaConfKey.READ_ONLY_VAR in options:
            if parse_boolean(options.get(SambaConfKey.READ_ONLY_VAR)) is False:
                self.read_only.disable()

        self.resource = resource
        # 3 options below are default
        # convert to string if needed as option value must be string
        self.directory_mask = SAMBA_DIR_ACCESS_MASK
        self.create_mask = SAMBA_DIR_ACCESS_MASK
        self.force_user = USER_NOBODY
        self.comment = options.get(SambaConfKey.COMMENT)
        if SambaConfKey.BROWSABLE in options and parse_boolean(options.get(SambaConfKey.BROWSABLE)):
            self.browsable = BooleanOption(SambaConfKey.BROWSABLE).enable()
        else:
            self.browsable = BooleanOption(SambaConfKey.BROWSABLE).disable()
        self.guest_ok = BooleanOption(SambaConfKey.GUEST_OK)
        if SambaConfKey.GUEST_OK_VAR in options:
            if parse_boolean(options.get(SambaConfKey.GUEST_OK_VAR)):
                self.guest_ok.enable()
        # coming from local configuration
        # or from service : can be a space or a underscore
        if SambaConfKey.GUEST_OK in options:
            if parse_boolean(options.get(SambaConfKey.GUEST_OK)):
                self.guest_ok.enable()

        self.strict_sync = BooleanOption(SambaConfKey.STRICT_SYNC_VAR)
        if SambaConfKey.STRICT_SYNC_VAR in options:
            if parse_boolean(options.get(SambaConfKey.STRICT_SYNC_VAR)):
                self.strict_sync.enable()
        # coming from local configuration
        # or from service : can be a space or a underscore
        if SambaConfKey.STRICT_SYNC in options:
            if parse_boolean(options.get(SambaConfKey.STRICT_SYNC)):
                self.strict_sync.enable()

        self.sync_always = BooleanOption(SambaConfKey.SYNC_ALWAYS_VAR)
        if SambaConfKey.SYNC_ALWAYS_VAR in options:
            if parse_boolean(options.get(SambaConfKey.SYNC_ALWAYS_VAR)):
                self.sync_always.enable()
        # coming from local configuration
        # or from service : can be a space or a underscore
        if SambaConfKey.SYNC_ALWAYS in options:
            if parse_boolean(options.get(SambaConfKey.SYNC_ALWAYS)):
                self.sync_always.enable()
        if self.sync_always.enabled() is True and self.strict_sync.enabled() is False:
            raise ValueError('Strict sync must be true when sync always is true')
        self.hosts_allow = options.get(SambaConfKey.HOSTS_ALLOW_VAR, options.get(SambaConfKey.HOSTS_ALLOW))
        self.hosts_deny = options.get(SambaConfKey.HOSTS_DENY_VAR, options.get(SambaConfKey.HOSTS_DENY))

    def __str__(self):
        '''
        String representation of the object
        '''
        out = StringIO.StringIO()
        out.write('[' + self.resource + ']')
        if self.comment:
            out.write('\n')
            out.write('%s = %s' % (SambaConfKey.COMMENT, self.comment))
        out.write('\n')
        out.write('%s = %s' % (SambaConfKey.READ_ONLY, self.read_only.enabled()))
        out.write('\n')
        out.write('%s = %s' % (SambaConfKey.GUEST_OK, self.guest_ok.enabled()))
        out.write('\n')
        out.write('%s = %s' % (SambaConfKey.STRICT_SYNC, self.strict_sync.enabled()))
        out.write('\n')
        out.write('%s = %s' % (SambaConfKey.SYNC_ALWAYS, self.sync_always.enabled()))
        out.write('\n')
        out.write('%s = %s' % (SambaConfKey.BROWSABLE, self.browsable.enabled()))
        if self.hosts_allow:
            out.write('\n')
            out.write('%s = %s' % (SambaConfKey.HOSTS_ALLOW, self.hosts_allow))
        if self.hosts_deny:
            out.write('\n')
            out.write('%s = %s ' % (SambaConfKey.HOSTS_DENY, self.hosts_deny))

        contents = out.getvalue()
        out.close()
        return contents

    def get_resource(self):
        '''
        Gets the resource name of the Samba Export
        '''
        return self.resource

    def get_id(self):
        '''
        Get the export ID i.e resource name
        '''
        return self.resource

    def get_options_as_dict(self):
        '''
        Get the options as dictionary item
        '''
        return dict({SambaConfKey.PATH: self.path,
                     SambaConfKey.COMMENT: self.comment,
                     SambaConfKey.BROWSABLE: self.browsable.enabled(),
                     SambaConfKey.READ_ONLY_VAR: self.read_only.enabled(),
                     SambaConfKey.GUEST_OK_VAR: self.guest_ok.enabled(),
                     SambaConfKey.STRICT_SYNC_VAR: self.strict_sync.enabled(),
                     SambaConfKey.SYNC_ALWAYS_VAR: self.sync_always.enabled(),
                     SambaConfKey.HOSTS_ALLOW_VAR: self.hosts_allow,
                     SambaConfKey.HOSTS_DENY_VAR: self.hosts_deny}
                    )

    def to_json(self):
        '''
        Represent the object as json.
        One use case of this function is to pass this as string while we send and retrieve data to/from
        priv-exec process
        '''
        return json.dumps({'type': self.type,
                           '__type__': self.__class__.__name__,
                           'id': self.get_id(),
                           SambaConfKey.PATH: self.path,
                           SambaConfKey.RESOURCE: self.resource,
                           'options': self.get_options_as_dict()
                           })

    def serialize(self):
        '''
        Serialize the object
        output: dictionary
        '''
        return dict(id=self.get_id(),
                    path=self.path,
                    resource=self.resource,
                    type=self.type,
                    __type__=self.__class__.__name__,
                    options=self.get_options_as_dict())

    @classmethod
    def deserialize(cls, js):
        '''
        Create an object from dictionary items
        input format:
            {
             'resource': 'test_smb_db',
             'type': 'smb',
             'options': {
                       'comment': 'blah blah',
                       'browsable': 'boolean',
                       'read_only': 'boolean',
                       'guest_ok': 'boolean',
                       'hosts_allow': 'x.x.x.x, y.y.y.y',
                       'hosts_deny': 'x.x.x.x, y.y.y.y'
                      }
            }
        '''
        js = json.loads(js)
        resource = js.get('resource')
        options = dict(js.get('options'))
        return cls(resource, **options)

    def __eq__(self, other):
        """
        Check for equality by testing resource name equality
        Resource name uniquely identify an entry
        """
        if id(self) == id(other):
            return True
        if other is None:
            return False

        if isinstance(other, SambaExportEntry):
            return self.get_id() == other.get_id()
        if isinstance(other, basestring):
            return self.__eq__(SambaExportEntry(other))

    def __ne__(self, other):
        """
        see __eq__
        """
        return not self.__eq__(other)


if __name__ == '__main__':
    set_log_config()
