"""
Systemd Service:
This modules handles different operations of systemd services like
Information retrieval, start, enable, stop etc.
Usage Example:
from system import SystemdService
service = SystemdService('httpd')
service.load_details()
service.start_service()
service.enable_service()
"""
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from backend.common import SUCCESS, SYSCTL, JOURNALCTL, DEFAULT_LOG_LINES
from backend.command_util import execute_command, run_command_no_output
from backend.log_util import get_logger, set_log_config


LOG = get_logger('lsa.systemdctl')


# TODO change this module name to SystemService or something similar
class SystemdService(object):
    START = "start"
    RESTART = "restart"
    STOP = "stop"
    ENABLE = "enable"
    DISABLE = "disable"
    SHOW = "show"
    RELOAD_OR_RESTART = "reload-or-restart"

    ALL_CMDS = [START, RESTART, STOP, ENABLE, DISABLE, SHOW, RELOAD_OR_RESTART]

    PROP_ID = "Id"
    PROP_NAMES = "Names"
    PROP_DESC = "Description"
    PROP_FRAGMENT_PATH = "FragmentPath"
    PROP_LOAD_STATE = "LoadState"
    PROP_ACTIVE_STATE = "ActiveState"
    PROP_SUB_STATE = "SubState"
    PROP_CGROUP = "ControlGroup"
    PROP_PID = "ExecMainPID"
    PROP_RESULT = "Result"

    LOADED_STATE = "loaded"

    def __init__(self, name):
        """
        we are using grep command to get some information for a service.
        adding .service to the end refines th output properly.
        """

        if not name:
            raise ValueError('name cannot be None nor empty')

        self.name = name
        self.is_installed = False
        self.logs = []
        self.cgroup = None
        self.code_stat = None
        self.code_out = None
        self.status = None
        self.active = None
        self.load = None
        self.substate = None
        self.vendor = None
        self.since = None
        self.s_file = None
        self.desc = None
        self.enabled = False
        self.resulti = None

    def _is_valid_command(self, command):
        if command not in SystemdService.ALL_CMDS:
            return False
        return True

    def _run_systemctl_command(self, service, command):
        """
        Execute a systemctl command and return SUCCESS/FAILURE
        service: service name
        command: type (start. stop, restart, status)
        """
        if not self._is_valid_command(command):
            raise KeyError("Not a valid command")
        try:
            run_command_no_output([SYSCTL, command, service])
        except Exception as e:
            raise Exception("Failed to execute systemctl command: " + str(e))

        return SUCCESS

    def load_details(self, include_logs=True):
        """
        Loads service attributes from runtime states
        parameters:
          include_logs: if True, logs are also fetched. the number of lines is limited to DEFAULT_LOG_LINES(=10000)
        returns: this instance
        raises:
            StantdardError: underneath command has failed
        """
        # we also ask for ID
        # this is not necessary as we already have the name
        # but that will be a test for non-installed service
        out, err, code = execute_command([SYSCTL,
                                          SystemdService.SHOW,
                                          self.name,
                                          '--no-pager',
                                          '-p',
                                          ','.join([SystemdService.PROP_ID,
                                                    SystemdService.PROP_DESC,
                                                    SystemdService.PROP_FRAGMENT_PATH,
                                                    SystemdService.PROP_LOAD_STATE,
                                                    SystemdService.PROP_ACTIVE_STATE,
                                                    SystemdService.PROP_SUB_STATE,
                                                    SystemdService.PROP_CGROUP,
                                                    SystemdService.PROP_PID,
                                                    SystemdService.PROP_RESULT])])

        if code != 0:
            raise StandardError('cannot get service details: %s' % err)
        for line in out.split('\n'):
            if line == '':
                continue
            key, value = line.split('=', 1)
            if key == SystemdService.PROP_ID:
                self.is_installed = True
            if key == SystemdService.PROP_DESC:
                self.desc = value
            if key == SystemdService.PROP_FRAGMENT_PATH:
                self.s_file = value
            if key == SystemdService.PROP_LOAD_STATE:
                self.load = value
            if key == SystemdService.PROP_ACTIVE_STATE:
                self.active = value
            if key == SystemdService.PROP_SUB_STATE:
                self.substate = value
            if key == SystemdService.PROP_PID:
                self.mainPid = value
            if key == SystemdService.PROP_CGROUP:
                self.cgroup = value
            if key == SystemdService.PROP_RESULT:
                self.result = value

        if self.load == SystemdService.LOADED_STATE:
            self.enabled = True

        # now retrieve logs.
        if include_logs is True:
            out, err, code = execute_command([JOURNALCTL,
                                              '--no-pager',
                                              '--lines=%s' % str(DEFAULT_LOG_LINES),
                                              '--unit=%s' % self.name,
                                              '--output=cat'])
            if code != 0:
                # we should not failed for only this
                LOG.warning('cannot get service logs: %s' % err)
            self.logs = out.split('\n')

        return self

    def serialize(self):
        """
        return serialized dictionary of this
        """
        return dict(__type__=self.__class__.__name__,
                    name=self.name,
                    desc=self.desc,
                    status=self.substate,
                    logs=self.logs,
                    result=self.result)

    def __str__(self):
        """
        Detailed string output of this class
        """
        out = ""
        if not self.is_installed:
            out += self.name + " is not installed."
            return out
        out += "Service name: " + self.name
        out += "\nDescription: " + self.desc
        out += "\nService file: " + self.s_file
        out += "\nEnable Status: %s" % self.enabled
        out += "\n Sub state: " + self.substate
        if self.enabled == "disabled":
            out += "\n" + self.active + "(" + self.load + ")"
            return out
        else:
            out += "\nLoad status:(" + self.load + ")"
        if self.vendor:
            out += "\n" + self.vendor
        out += "\n" + self.active + "(" + self.substate + ")"
        if self.since:
            out += "\nSince: " + self.since
        if self.cgroup:
            out += "\nCGroup: " + self.cgroup
        if self.code_stat:
            if self.status == "running":
                out += "\nCode File: " + self.code_stat
            else:
                out += "\n" + self.code_stat
        if self.code_out:
            if self.status == "running":
                out += "\n code file: " + self.code_out
            else:
                out += "\n" + self.code_out
        if self.logs:
            for log in self.logs:
                out += "\n" + log
        return out

    def enable_service(self):
        """
        Enable the service to automatically start after boot
        Parameter:
        Input: none
        Output: Success or raise exception
        """
        if not self.is_installed:
            LOG.error("Trying to enable uninstalled service " + self.name)
            raise Exception("Trying to enable uninstalled service " + self.name)
        return self._run_systemctl_command(self.name, SystemdService.ENABLE)

    def disable_service(self, stop_it=False):
        """
        Disable the service
        Parameter:
        Input:
          stop_it : if True the service is also stopped
        Output: Success or raise exception
        """
        if not self.is_installed:
            LOG.error("Trying to disable uninstalled service " + self.name)
            raise Exception("Trying to disable uninstalled service " + self.name)
        cmd = [SYSCTL, SystemdService.DISABLE]
        if stop_it:
            cmd.append('--now')
        cmd.append(self.name)
        try:
            run_command_no_output(cmd)
        except Exception as e:
            raise Exception("Failed to execute systemctl command: " + str(e))

    def start_service(self):
        """
        Start the service
        Parameter:
        Input: none
        Output: Success or raise exception
        """
        if not self.is_installed:
            LOG.error("Trying to start uninstalled service " + self.name)
            raise Exception("Trying to enable uninstalled service " + self.name)
        return self._run_systemctl_command(self.name, SystemdService.START)

    def stop_service(self):
        """
        Stop the service
        Parameter:
        Input: none
        Output: Success or raise exception
        """
        if not self.is_installed:
            LOG.error("Trying to stop uninstalled service " + self.name)
            raise Exception("Trying to enable uninstalled service " + self.name)
        return self._run_systemctl_command(self.name, SystemdService.STOP)

    def restart_service(self):
        """
        Restart the service
        Parameter:
        Input: none
        Output: Success or raise exception
        """
        if not self.is_installed:
            LOG.error("Trying to restart uninstalled service " + self.name)
            raise Exception("Trying to enable uninstalled service " + self.name)
        return self._run_systemctl_command(self.name, SystemdService.RESTART)


def run_all_test():
    print "--Filewalld----"
    sys = SystemdService("firewalld")
    sys.load_details()
    print str(sys)
    print "Jason---\n"
    print sys.toJSON()

    print "-----------------target---"
    sys = SystemdService("target").load_details()
    print "Jason---\n"
    print sys.toJSON()

    print str(sys)
    print "-----------------ceph-mon---"
    sys = SystemdService("ceph-mon").load_details()
    print str(sys)
    print "--------nfs-idmapd---"
    sys = SystemdService("nfs-idmapd").load_details()
    print str(sys)
    print "----- nfs-mountd----"
    sys = SystemdService("nfs-mountd").load_details()
    print str(sys)
    print "----- httpd----"
    sys = SystemdService("httpd").load_details()
    print str(sys)
    print "----- tgtd----"
    sys = SystemdService("tgtd").load_details()
    print str(sys)


def run_command_test():
    print "---starting comamnd test---"
    sys = SystemdService("target")
    print str(sys)
    print "----starting service---"
    sys.start_service()
    print str(sys)
    print "----stopping service---"
    sys.stop_service()
    print str(sys)
    print "----restarting service---"
    sys.restart_service()
    print str(sys)


if __name__ == '__main__':
    set_log_config()
    # run_command_test()
    run_all_test()
