"""
Long running task framework package

This module hosts different types related to the framework

"""
__copyright__ = "Copyright 2017 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl."


from enum import Enum
import time
import random
import copy
import threading
from task_manager import TaskManager


__all__ = ['set_task_manager', 'get_task_manager',
           'TaskState', 'TaskStatus', 'TaskEvent', 'Task']

# global reference for main task manager
__task_manager = None


def get_task_manager():
    """
    Gets the global/common reference on task manager
    Returns a TaskManager instance
    """
    global __task_manager
    if __task_manager is None:
        raise StandardError('task manager not initialized')

    return __task_manager


def set_task_manager(task_manager):
    """
    Sets the global/common reference on task manager
    parameter:
       task_manager:  a TaskManager instance
    """
    global __task_manager
    if not isinstance(task_manager, TaskManager):
        raise TypeError('not a task manager')
    if __task_manager is not None:
        raise StandardError('task manager already initialized')

    __task_manager = task_manager


class TaskState(Enum):
    # Task has been created but not yet started nor aborted
    CREATED = 0
    # Task is starting. between STARTING and START denote task initialisation time
    STARTING = 1
    # Task is started. a task is started as soon as it has started doing actual operations
    # as opposed to STARTING
    STARTED = 2
    # Task is stopping
    STOPPING = 3
    # Task is aborting
    ABORTING = 4
    # Task is stopped
    STOPPED = 5
    # Task is aborted
    ABORTED = 6

    def serialize(self):
        """
        Serialise this for JSON
        Returns a JSON string of this object
        """
        return self.name


class TaskStatusCode(Enum):
    """
    Task (operation) status code
    """
    SUCCESS = 0
    FAILED = 1


class TaskStatus(object):
    """
    Task operation status
    """
    def __init__(self, code, message=''):
        """
        Instantiates a new TaskStatus
        paremeters:
          code: status code, one of TaskStatusCode
          message: optional message associated with the status
        """
        if not isinstance(code, TaskStatusCode):
            raise ValueError('Invalid status code %s' % code)

        self.code = code

        self.message = message

    def serialize(self):
        """
        Serialize this for JSON
        Returns a JSON string of this object
        """
        # __type__ is used by deserializer
        return dict(__type__=self.__class__.__name__,
                    code=self.code.value,
                    message=self.message)

    def __str__(self):
        return "TaskStatus {code=%s, message=%s}" % (self.code.name, self.message)

    @classmethod
    def deserialize(cls, data):
        """
        Creates an object from (JSON) data as dictionnary
        Returns a new TaskStatus instance
        """
        code = data.get('code')
        if code is None:
            raise KeyError('missing \'code\' key')
        try:
            code = TaskStatusCode(code)
        except KeyError:
            raise KeyError('Invalide code %d' % code)
        msg = data.get('message')
        if msg is None:
            raise KeyError('missing \'message\' key')

        return cls(code, msg)


class TaskStateChange(object):
    """
    Task lifecycle state.
    Records a transition between two TaskState
    """

    def __init__(self, state_type, message=''):
        """
        Instantiates a new TaskStateChange
        paremeters:
          state_type: task state, one of TaskState
          message: optional message associated with the status
        """
        if not isinstance(state_type, TaskState):
            raise TypeError('invalid state change type')

        self.state_type = state_type
        self.time = time.time()
        self.message = message

    def __str__(self):
        return "TaskStateChange {type=%s, time=%s, message=%s}" % (self.state_type.name,
                                                                   time.ctime(self.time),
                                                                   self.message)

    def __eq__(self, other):
        """
        check for equality
        if other is of type TaskState, check that this type match
        otherwsie check for type and time equality
        """
        if isinstance(other, TaskState):
            return self.state_type == other
        if isinstance(other, TaskStateChange):
            return self.state_type == other.state_type and self.time == other.time
        raise TypeError('cannot compare against %s' % type(other))

    def __lt__(self, other):
        """
        Checks if this is less than other by comparing timestamp
        if other is of type TaskState, comparison is made about the state
        ex:
          TaskStateChange(...) < TaskState.STOPPED
        """
        if isinstance(other, TaskState):
            return self.state_type.value < other.value
        if isinstance(other, TaskStateChange):
            if self.time != other.time:
                return self.time < other.time
            return self.state_type.value < other.state_type.value
        raise TypeError('cannot compare against %s' % type(other))

    def __gt__(self, other):
        """
        Checks if this is greater than other by comparing timestamp
        if other is of type TaskState, comparison is made about the state
        ex:
          TaskStateChange(...) > TaskState.STOPPED
        """
        if isinstance(other, TaskState):
            return self.state_type.value > other.value
        if isinstance(other, TaskStateChange):
            if self.time != other.time:
                return self.time > other.time
            return self.state_type.value > other.state_type.value
        raise TypeError('cannot compare against %s' % type(other))

    def get_time(self):
        """
        Return the time in seconds from EPOC of that state change
        """
        return self.time

    def serialize(self):
        """
        creates an object from (JSON) data as dictionnary
        returns a dict
        """
        # __type__ is used by deserializer
        return dict(__type__=self.__class__.__name__,
                    type=self.state_type,
                    time=self.time,
                    message=self.message)

    @classmethod
    def deserialize(cls, data):
        """
        Creates an object from (JSON) data as dictionnary
        Returns a new TaskStateChange instance
        """
        msg = data.get('message')
        if msg is None:

            raise KeyError('missing \'message\' key')
        type_str = data.get('type')
        if not type_str:
            raise KeyError('missing \'type_str\' key')
        time_str = data.get('time')
        if not time_str:
            raise KeyError('missing \'time_str\' key')

        state = None
        try:
            state = TaskState[type_str]
        except KeyError, e:
            raise KeyError('unexpected state change %s' % str(e))
        time = None
        try:
            time = float(time_str)
        except KeyError, e:
            raise KeyError('unexpected state change %s' % str(e))
        _n_state_change = cls(state, msg)
        _n_state_change.time = time
        return _n_state_change


class Task:
    """
    Long running task super class
    To be subclassed by other long running task implementors
    ex:
      FooTask(Task):
        def begin(self):
           Task.begin(self)
           #... perform some init
           self.states.append(TaskStateChange(TaskState.STARTED,'this is started'))
           #... task logic starts here
        def end(self):
           Task.end(self)
           # ... stop things
           self.states.append(TaskStateChange(TaskState.STOPPED, 'this stopped'))
    Another use of Task object is to share a synchronization mechanism
    i.e task's method actually do nothing but to maintain state change of a long running
    operation. I.e logic is kept in other object and Task instance is just there to kept
    trackof progression
    """

    __states_protection = threading.Lock()

    # init random generator for futur Task instance creations
    random.seed()

    def __init__(self, target, name=None):
        """
        Instantiates a new task
        parameter
         target: the target of this task,  it is related to. This is usually a share name
         name : optional name of that new task
        """
        if not target:
            raise ValueError('Invalid name property')
        self.target = target

        self.name = name
        # we do not expect collision as we do trash task repository
        # time to time.
        self.ID = random.randint(0, 100000)
        self.states = [TaskStateChange(TaskState.CREATED)]
        self.status = None

    def __str__(self):
        return "Task {target=%s, name=%s, ID=%d, status=%s}" % (self.target, self.name, self.ID, self.status)

    def __eq__(self, other):
        """
        Overwrite eq operator by comparing task's id
        """
        if not isinstance(other, Task):
            raise TypeError('cannot compare against %s' % type(other))
        return self.ID == other.ID

    @staticmethod
    def __get_create_time(obj):
        """
        Note : There is always a creation time
        """
        return [state_change.get_time() for state_change in obj.states if state_change == TaskState.CREATED][0]

    def __gt__(self, other):
        """
        compare creation date of tasks
        """
        create_time = Task.__get_create_time(self)
        if isinstance(other, int):
            return create_time > other
        if isinstance(other, Task):
            other_create_time = Task.__get_create_time(other)
            return create_time > other_create_time
        raise TypeError('cannot compare against %s' % type(other))

    def __ge__(self, other):
        """
        compare creation date of tasks
        """
        create_time = Task.__get_create_time(self)
        if isinstance(other, int):
            return create_time >= other
        if isinstance(other, Task):
            other_create_time = Task.__get_create_time(other)
            return create_time >= other_create_time
        raise TypeError('cannot compare against %s' % type(other))

    def __lt__(self, other):
        """
        compare creation date of tasks
        """
        create_time = Task.__get_create_time(self)
        if isinstance(other, int):
            return create_time < other
        if isinstance(other, Task):
            other_create_time = Task.__get_create_time(other)
            return create_time < other_create_time
        raise TypeError('cannot compare against %s' % type(other))

    def __le__(self, other):
        """
        compare creation date of tasks
        """
        create_time = Task.__get_create_time(self)
        if isinstance(other, int):
            return create_time <= other
        if isinstance(other, Task):
            other_create_time = Task.__get_create_time(other)
            return create_time <= other_create_time
        raise TypeError('cannot compare against %s' % type(other))

    @classmethod
    def deserialize(cls, data):
        """
        Creates an object from (JSON) data as dictionnary
        Returns a new Task instance
        """
        target = data.get('target')
        if target is None:
            raise KeyError('missing \'target\' key')

        name = data.get('name')

        status = data.get('status')
        if not status:
            raise KeyError('missing \'status\' key')

        ID = data.get('ID')
        if ID is None:
            raise KeyError('missing \'ID\' key')

        state_changes_list = data.get('states')
        if state_changes_list is None:
            raise KeyError('missing \'states\' key')

        new_task = cls(target, name)
        new_task.ID = ID
        new_task.states = state_changes_list
        new_task.status = status
        return new_task

    def get_target(self):
        """
        Gets the target of this operation
        Returns:
          target as string
        """
        return self.target

    def get_name(self):
        """
        Gets the name of this task
        Returns:
          name as string
        """
        return self.name

    def get_status(self):
        """
        Gets the status of the task
        Return the status as TaskStatus or None if the Task as actually
        no status (operation not completed)
        """
        return self.status

    def get_state(self):
        """
        Returns the current state of the task
        as TaskStateChange
        """
        return max(self.states)

    def get_states(self):
        """
        Returns the states of the task
        as list of TaskStateChange
        """
        return self.states

    def is_running(self):
        """
        Return True if the tas is actually running
        """
        state = self.get_state()
        return state > TaskState.CREATED and state < TaskState.STOPPED

    def get_id(self):
        """
        Gets the task id as int
        """
        return self.ID

    def serialize(self):
        """
        Creates an object from (JSON) data as dictionnary
        returns a dict
        """
        # cannot rely on class name as we may be subclassed
        # use 'Task' instead
        # __type__ is used by deserializer
        with Task.__states_protection:
            _s = dict(__type__="Task",
                      target=self.target,
                      name=self.name,
                      ID=self.ID,
                      states=copy.deepcopy(self.states),
                      status=copy.deepcopy(self.status))
        return _s

    def _add_state(self, state_change):
        """
        Private method for this class and sub classes
        """
        with Task.__states_protection:
            self.states.append(state_change)

    def begin(self):
        """
        super class begin method which has to be overwritten
        by implementors. Subclass must call it first
        """
        self._add_state(TaskStateChange(TaskState.STARTING))

    def end(self, status):
        """
        super class end method which has to be overwritten
        by implementors
        This method is called by this task once operation is over
        Subclass must call it first
        parameter:
           status: operation status as TaskStatus
        """
        self._add_state(TaskStateChange(TaskState.STOPPING))
        with Task.__states_protection:
            self.status = status

    def abort(self):
        """
        abort method which has to be overwritten
        by implementors. As opposed to the method 'end'
        This is intended to be called by the task manager
        Subclass must call it first
        """
        self._add_state(TaskStateChange(TaskState.ABORTING))
