"""
Utility functions(formatting numbers, size and others which can be used by
different modules.
"""
__copyright__ = """
Copyright 2017 Oracle.  All rights reserved.
Author: Muminul Islam <muminul.islam@oracle.com>
"""
import re
import os
from backend.common import SHARE_POINT, ROOT_DEV, VG_NAME


def format_readable_to_bytes(readable_size, kilo=1024):
    '''
    Convert readable size to bytes.
      parameter:
        readable_size: string, example (10kb. 10kib, xxgb etc)
        output: size in bytes in integer
    '''
    # replace i with blank just in case anyone provides Mib GiB or so
    size = readable_size.replace('i', '')
    size = readable_size.lower()
    if not re.match('(\d+(?:\.\d+)?)([g|t|k|m]+[b]*)$', size, re.IGNORECASE):
        raise ValueError("Cannot interpret size, wrong format: %s" % readable_size)
    size = size.rstrip('b')

    units = ['k', 'm', 'g', 't']
    try:
        power = units.index(size[-1]) + 1
    except ValueError:
        power = 0
        size = float(size)
    else:
        size = float(size[:-1])
    powered = int(kilo) ** power
    size_in_bytes = size * float(powered)
    return int(size_in_bytes)


def format_bytes_to_readable(size):
    '''
    covert bytes to readable
      parameter:
          size: bytes
          output: readable size as string (xxxGB, xxxTB etc)
    '''
    kilo = float(1024.0)
    size = float(size)
    # don't use decimal for bytes
    if size < kilo:
        return "%d bytes" % size
    size = size / kilo

    for x in ['KB', 'MB', 'GB', 'TB']:
        if size < kilo:
            break
        size = size / kilo
    if size.is_integer():
        size = int(size)
    else:
        size = round(size, 2)
    return "%s%s" % (size, x)


def format_human_readable(readable_size):
    '''
    Format readable size to generic format. We show user on GB/MB/TB/KB format.
      parameter:
          readable_size: string format( g, G, Gib, gb, TB etc)
          output: formatted readable size ( GB, MB, TB, KB)
                  Raise exception for wrong format input
    '''
    size = readable_size.replace('i', '')
    size = size.lower()
    match_obj = re.match('(\d+(?:\.\d+)?)([g|t|k|m]+[b]*)$', size, re.IGNORECASE)
    if not match_obj:
        raise ValueError("Cannot interpret size, wrong format: %s" % readable_size)
    num = float(match_obj.groups()[0])
    # format float 1.00 as 1 1.006 as 1.01 etc
    if num.is_integer():
        num = int(num)
    else:
        num = round(num, 2)
    suffix = match_obj.groups()[1].rstrip('b')
    size = str(num) + suffix
    return size.upper() + 'B'


def get_full_mount_path(name):
    '''
    Get mount point of the share
      parameter:
        name: share/lv name
        output: mount point
    '''
    return os.path.join(os.path.sep, SHARE_POINT, name)


def get_full_lv_path(name):
    '''
    Get mount point of the share
      parameter:
        name: share/lv name
        output: lv patch ( soft link)
     '''
    return os.path.join(os.path.sep, ROOT_DEV, VG_NAME, name)


def get_proc_mount_lv_path(name):
    '''
    Get mount point of the share, typically /dev/mapper/lsa-name
      parameter:
        name: share/lv name
        output: mount point of share (actual path)
    '''
    return os.path.join(os.path.sep, ROOT_DEV, 'mapper', VG_NAME + '-' + name)


def parse_boolean(value):
    '''
    Parse the boolean  value:
    input might be Boolean, String or integer
    output: True/False
    '''
    if value is None:
        raise ValueError('Cannot parse boolean: value can not be None')
    val_lower = str(value).lower()
    if val_lower in ['1', 'true', 'yes']:
        return True
    elif val_lower in ['0', 'false', 'no']:
        return False
    else:
        raise ValueError('Invalid value: %s' % val_lower)


def round_to_multiplier(value, divider):
    '''
    Returns the closest value that can be divised by divider without remainder
    parameter:
         value: value to be rounded
         divider: divider to be used for the rounding
         output: rounded integer
    '''
    assert (divider != 0)
    return int(divider * round(float(value) / divider))
