""" backend file used for testing to simulate real backend calls """
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
import psutil
import cStringIO
from subprocess import Popen, PIPE
from flask.ext.login import current_user
import logging
import json
from backend.nfs_util import NfsExportEntry
import priv_exec_server as priv_exec
import backend.api_wrapper
from backend.common import CMD_HOSTNAME, CMD_NETSTAT
from backend.common import CMD_UNAME, ORACLE_RELEASE_FILE
from backend.common import (LSA_STATE_UNCONFIGURED, LSA_STATE_REMOVE, LSA_STATE_CONFIGURED, LSA_STATE_PENDING, LSA_STATE_READY)
import backend.command_util
from backend.log_util import set_log_config
from backend.oci_config_util import OCIInstanceHelper
from backend.share_util import ShareEntry
from backend.samba_util import SambaExportEntry
from backend.storage_pool import StoragePool

# use a child of main app logger: so leaf of 'lsa'
# logging configuration is done at 'lsa' logger level
logger = logging.getLogger('lsa.backend_test')


def get_hostname():
    """
    Get the hostname by linux hostname command
    """
    process = Popen([CMD_HOSTNAME], stdin=None, stdout=PIPE, stderr=None)
    out, err = process.communicate()
    if process.returncode:
        return "Hostname error"
    if out:
        return out
    return "Hostname error"


def get_host_ip_address():
    """
    Get the host IP address. This function always returns the IP address from
    first interface.
    """
    process = Popen([CMD_HOSTNAME, '-I'], stdin=None, stdout=PIPE, stderr=None)
    out, err = process.communicate()
    if process.returncode:
        return "Host IP address error"
    if out:
        return out.split()[0]
    return "Host IP address error"


def get_kernel_version():
    """
    Get the kernel version by linux `uname -r` command
    Output: kernel version. eg '4.1.12-61.1.18.el7uek.x86_64'
    Raise: Standard error if fails
    """
    cmdline = CMD_UNAME + " -r"
    out = None
    try:
        out = backend.command_util.output_shell_command(cmdline)
    except Exception as e:
        raise StandardError('Failed to get kernel version: %s' % str(e))
    return out.replace('\n', '')


def get_os_version():
    """
    Get the os version by linux `cat /etc/oracle-release` command
    Output: os version. eg 'Oracle Linux Server release 7.3'
    Raise: IOError,OSError if fails
    """
    try:
        fd = open(ORACLE_RELEASE_FILE, 'r')
        data = fd.read().replace('\n', '')
    except (IOError, OSError) as e:
        raise StandardError("Failed to get os version from file %s: %s" % (ORACLE_RELEASE_FILE, str(e)))
    finally:
        if fd:
            fd.close()
    return data


def get_platform_version():
    """
    Get the platform version by `rpm -q --qf '%{VERSION}' linux-storage-manager`
    Output: platform version. eg '0.0.3'
    Raise: Standard error if fails
    """
    result = priv_exec.priv_exec('get_platform_version',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['content']


def get_platform_disks():
    """
    In the current phase, list all of the iscsi disks.
    The info of the disks are from output of 'iscsiadm' and 'lsblk'. oci_utils already
    parse the output. This funtion leverages the oci_utils. This implies the ocid-service
    should be running when this function is called
    Output: list of iscsi disks
    Raise: Standard error if call to oci_utils fails
    """
    logger.debug('current_user [%s]' % str(current_user))
    logger.debug('current_user.user_key [%s]' % str(current_user.user_key))
    result = priv_exec.priv_exec('get_platform_disks',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['disks']


def systemUsages():
    usage = {'cpu': psutil.cpu_percent(),
             'mem': psutil.virtual_memory().percent,
             'mem_avail': psutil.virtual_memory().available,
             'swap': psutil.swap_memory().percent,
             'disk': psutil.disk_usage('/').percent,
             'disk_avail': psutil.disk_usage('/').free,
             'bytes_sent': psutil.net_io_counters().bytes_sent,
             'bytes_recv': psutil.net_io_counters().bytes_recv}
    return usage


def get_ksplice_autoinstall_state():
    '''
    Get the current setting of ksplice autoupdate
      Output: yes/no
      Raise: Standard error if fails
    '''
    result = priv_exec.priv_exec('get_ksplice_autoinstall_state',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['autoinstall_state']


def get_appliance_configuration_state():
    '''
    Get the appliance configuration status
    returns string CONFIGURED or UNCONFIGURED
    raise: StandardError if fails
    '''
    result = priv_exec.priv_exec('get_appliance_configuration_state',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['state']


def toggle_ksplice_autoupdate_setting(cur_state):
    '''
    Toggle the ksplice autoupdate settings (yes-->no, no-->yes)
    param:
        cur_sate: current setting of ksplice autoupdate
    raise: standard error in case of failure
    '''
    result = priv_exec.priv_exec('toggle_ksplice_autoupdate_setting',
                                 {'cur_state': cur_state,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['autoinstall_state']


def get_vg_stats():
    '''
    Get volume group usage
    output: dictionary {'vg_size': 'xxxx', 'vg_free': 'xxxx', 'status': 'active/faulty'}
    raise: standard error in case of failure
    '''
    result = priv_exec.priv_exec('get_vg_stats',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    json_obj = json.loads(result['vg_stats'])
    return {'vg_size': json_obj['vg_size'], 'vg_free': json_obj['vg_free'], 'status': json_obj['status']}


def update_ssh_keys(ssh_pubkeys):
    '''
    Add a ssh key to the server's authorized list
    Parameter:
      ssh_pubkeys: ssh public key as string
      Output: Nothing on success
      raise: Standard error if fails
    '''
    result = priv_exec.priv_exec('update_ssh_keys',
                                 {'ssh_keys': ssh_pubkeys,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def get_ssh_keys():
    result = priv_exec.priv_exec('get_ssh_keys',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['ssh_keys']


def set_samba_options(data):
    '''
    Update global samba options
    Parameter:
      options: dict which includes options['workgroup'], options['netbios name'], and options['interfaces']
      Output: Nothing on success
      raise: Standard error if fails
    '''
    options = json.loads(data)
    result = priv_exec.priv_exec('set_samba_options',
                                 {'options': options,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def get_samba_options():
    '''
    Get global samba options
    example:
    options['workgroup'] = 'workgroup_test'
    options['netbios name'] = 'netbios_test'
    options['interfaces'] =
    '''
    result = priv_exec.priv_exec('get_samba_options',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['options']


def get_instance_details():
    '''
    get oci instance details
    Output JSON format:
        {'region': location of the instance(string)
         'domain': availability domain(string)
         'shape': type of shape(string)
         'instance_name': instance/display name(string)
         'creation_time': time when instance created(string)
        }
    raise: Standard error on failure
    '''

    result = priv_exec.priv_exec('get_instance_details',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['instance_details']


def get_storage_pool():
    """
    Internal method to retrieve the storage pool
    raise:
      LookupError: the pool is not available in the host.
    """
    try:
        result = priv_exec.priv_exec('get_storage_pool',
                                     {'user_key': current_user.user_key})
        return StoragePool.deserialize(result['pool'])
    except Exception as e:
        raise LookupError('pool unavailable: %s' % str(e))


def storageUsages():
    """
    Returns as mountpoint indexed dict usages as percentage
    ex: {'/': 12,'data': 45}
    """
    sysRootUsagePct = get_mountpoint_usage('/').percent
    pool = None
    try:
        pool = get_storage_pool()
    except LookupError:
        logger.exception('pool not available')
        # go on anyway as we return info about root FS
    if pool:
        vg_size = float(pool.get_size())
        vg_free = float(pool.get_free())
        poolMntUsagePct = (((vg_size - vg_free) / vg_size) * 100.0)
        return {'/': sysRootUsagePct,
                'data': poolMntUsagePct}
    return {'/': sysRootUsagePct}


def get_mountpoint_usage(mntp='/'):
    return backend.api_wrapper.StorageManager.get_filesystem_usage(mntp)


def get_pool_usage():
    """
    gets the usages of the pool
    return total and free space as dictionary
    {'total': 123123, 'free':123123}
    raise LookupError: when pool is not available
    """
    try:
        pool = get_storage_pool()
    except Exception as e:
        raise LookupError('pool unavailable: %s' % str(e))

    return {'total': pool.get_size(), 'free': pool.get_free(), 'state': pool.get_status()}


def get_pool_status():
    """
    return pool status information as RaidArray instance
    see raid_util.RaidArray
    raise LookupError: when pool is not available
    """
    try:
        return get_storage_pool()
    except Exception as e:
        raise LookupError('pool unavailable: %s' % str(e))


def set_appliance_state(state):
    """
    Sets the appliance state
    Param:
      state : appliance state as in. one of the value LSA_STATE_*
    raise KeyError : for invalid state
    raise StandardError: error raised during state change
    """
    if state not in [LSA_STATE_UNCONFIGURED, LSA_STATE_REMOVE, LSA_STATE_CONFIGURED,
                     LSA_STATE_PENDING, LSA_STATE_READY]:
        raise KeyError('unknown type %s' % str(state))
    result = priv_exec.priv_exec('set_appliance_state',
                                 {
                                     'state': str(state),
                                     'user_key': current_user.user_key
                                 })
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def create_pool(raid_level, devices):
    """
    Call priv_exec server to create the storage pool from a list of devices
    expected argument:
        raid_level: raid level (0, 1, 6, 10, or RAID_LEVEL_NONE)
        devices: list of comma-separated paths (e.g., '/dev/sdb,/dev/sdc')
    raise:
         StandardError: error raised during creation
    """
    result = priv_exec.priv_exec('create_pool',
                                 {'raid_level': raid_level,
                                  'devices': devices,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


_allServiceNames = ["nfs-idmapd",
                    "nfs-mountd",
                    "nfs-server",
                    "smb",
                    "nmb",
                    "httpd"]


def getServices():
    """
    retrieve services status
    services taken into account are the ones
    important for appliance good health
     nfs-idmapd,nfs-mountd,nfs-server,httpd
    raise StandardError: underneath operation has failed
    return an array of service as JSON object
    """
    # TODO: avoid serialize/deserialize JSON and str twice...
    result = priv_exec.priv_exec('get_services_status',
                                 {'service_names': ','.join(_allServiceNames),
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    # priv backend send us array of string
    # jsonify back all instance
    return [json.loads(svcAsString) for svcAsString in result['status']]


def get_shares(names=[]):
    """
    Gets share list
    parameter:
        names: list of names  of the share we want to retrieve
    Return a list of dictionary
    Each share has the following
      {
             'name':the share name,
             'size':share size in bytes as int,
             'free': free space in bytes as int,
             'fs_options': {'inode32': false} filesystem options
      }
    """
    result = priv_exec.priv_exec('list_shares',
                                 {'names': names,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    try:
        deserializedEntries = [ShareEntry.deserialize(share) for share in result['shares']]
    except Exception as e:
        raise StandardError(str(e))
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("Shares returned")
        for entry in deserializedEntries:
            logger.debug(entry)

    return [{'name': share.get_name(),
             'size': share.get_size(),
             'free': share.get_free(),
             'fs_options': {'inode32': share.get_inode32()}
             }
            for share in deserializedEntries]


def delete_nfs_export(name, hosts):
    """
    Call priv_exec server to delete a new NFS share
    expected argument:
        'name': The name for the new NFS share
        'hosts': hosts the share is exported to
    raise:
         StandardError: error raised during creation
    """
    result = priv_exec.priv_exec('delete_nfs_export',
                                 {'name': name,
                                  'hosts': hosts,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def add_nfs_export(name, hosts, options):
    """
    Call priv_exec server to create a new NFS share
    expected argument:
        'name': The name for nfs share/directory
        'hosts': hostname / ip address
        'options': the share options as expected by NFS
    returns: Share information as  dictionary
    raise:
         StandardError: error raised during creation
    """
    result = priv_exec.priv_exec('create_nfs_export',
                                 {'name': name,
                                  'hosts': hosts,
                                  'options': options,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    entry = NfsExportEntry.deserialize(result['entry'])
    return {"id": entry.get_id(), "path": entry.get_path(),
            "hosts": entry.get_hosts(),
            "options": entry.get_options_as_dict()}


def modify_nfs_export(name, hosts, options):
    """
    Call priv_exec server to modify an NFS share
    expected argument:
        'name': The name for nfs share/directory
        'hosts: authorized client for the nfs export
        'options': the share options as expected by NFS
    returns: return share information as a dictionary
    raise:
         StandardError: error raised during creation
    """
    result = priv_exec.priv_exec('modify_nfs_export',
                                 {'name': name,
                                  'hosts': hosts,
                                  'options': options,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    entry = NfsExportEntry.deserialize(result['entry'])
    return {'id': entry.get_id()}


def create_samba_export(name, resource, options):
    '''
    Call priv_exec to create a samba export
    expected arguments:
        'name': name of the share
        'resource': resource name of the samba export
        'options': samba export options as dictionary items. Format example
        {"comment": "comment", "browsable": "boolean", "guest_ok": "boolean", "strict_sync": "boolean", "sync_always": "boolean", "hosts_allow": "x.x.x.x",
             "hosts_deny": "x.x.x.x", "read_only": "boolean"}
    raise: StandardError
    '''
    result = priv_exec.priv_exec('create_samba_export',
                                 {'name': name,
                                  'resource': resource,
                                  'options': options,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    entry = SambaExportEntry.deserialize(result['entry'])
    return {"id": entry.get_id()}


def modify_samba_export(name, resource, options):
    '''
    Call priv_exec to create a samba export
    expected arguments:
        'name': name of the share
        'resource': resource name of the samba export
        'options': samba export options as dictionary items. Format example
                {"comment": "comment", "browsable": "boolean", "guest_ok": "boolean", "hosts_allow": "x.x.x.x",
                 "strict_sync": "boolean", "sync_always": "boolean", "hosts_deny": "x.x.x.x", "read_only": "boolean"}
    raise: StandardError for failure
    output: export id and resource as dictionary items example: {'id': 'some_id', 'export': 'export_name'}
            For samba id and export are the same.
    '''
    result = priv_exec.priv_exec('modify_samba_export',
                                 {'name': name,
                                  'resource': resource,
                                  'options': options,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    entry = SambaExportEntry.deserialize(result['entry'])
    return {"id": entry.get_id(), "resource": entry.resource}


def delete_samba_export(name, resource):
    '''
    Call priv_exec to delete a samba export
    expected arguments:
        'name' : share name
        'resource': resource name of the samba export
    raise: StadanrdError
    '''
    result = priv_exec.priv_exec('delete_samba_export',
                                 {'resource': resource,
                                  'name': name,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def create_share(name, size, inode32):
    '''
    Create a share in backend storage
      parameter:
        name: name of the share
        size: Size of the nfs share in bytes
        inode32: workaround for legacy application limitation for 26879349
    returns: dictionary as {'task_id':<task id>}
    raise:
          StandardError: error raised during creation
    '''
    result = priv_exec.priv_exec('create_share',
                                 {'name': name,
                                  'size': size,
                                  'inode32': inode32,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['task_id']


def resize_share(name, size):
    '''
    Resize a share in backend storage
      parameter:
        name: name of the share
        size: Size of the share in bytes
    raise:
          StandardError: error raised during modification
    '''
    result = priv_exec.priv_exec('resize_share',
                                 {'name': name,
                                  'size': size,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def delete_share(name):
    '''
    Delete a share in back storage
      parameter:
        name: name of the share as string
    returns: nothing
    raise:
          StandardError
    '''
    result = priv_exec.priv_exec('delete_share',
                                 {'name': name,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def restartNFSStack():
    """
    Calls priv_exec server to restart all NFS related services in order
    raise:
        StandardError: error raised during restart
    """
    result = priv_exec.priv_exec('admin_restart_nfs',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def restartSambaServices():
    """
    Calls priv_exec server to restart all Samba related services in order
    raise:
        StandardError: error raised during restart
    """
    result = priv_exec.priv_exec('admin_restart_smb',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


def get_logs(logType, keyword):
    result = priv_exec.priv_exec('get_logs',
                                 {'log_type': logType,
                                  'keyword': keyword,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result['content']


def get_nfs_clients():
    '''
    returns a list of ip addresses corresponding to established connections
    to the NFS server assuming port 2049, None is errors
    can be an empty list
    '''
    sysip = get_host_ip_address()
    sysip += ':2049'
    raw_res = Popen([CMD_NETSTAT, '-an'], stdin=None, stdout=PIPE, stderr=None).communicate()[0].split('\n')
    res = []
    for i in raw_res:
        if sysip in i:
            # only append client ip. the input format expected is:
            # where 10.166.175.251 is the nfs client
            # 'tcp        0      0 10.166.175.73:2049      10.166.175.251:996  ESTABLISHED'
            res.append(i.split()[4].split(':')[0])

    return res


def system_reboot():
    result = priv_exec.priv_exec(
        'reboot',
        {'user_key': current_user.user_key}
    )
    # we ignore error coming from broken pipe case as we are triggering a reboot
    if result['results'] == 'failure' and result['code'] != 2:
        raise StandardError(result['reason'])


def system_update():
    # 10 min timeout as the update can be a long running process
    result = priv_exec.priv_exec(
        'update',
        {'user_key': current_user.user_key},
        600)
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result


def backup_share_on_object_store(share_name, namespace_name, bucket_name, obj_name):
    """
    Creates a backup of a share.
    The content of the share is dumped and then pushed to remote object store
    arguments :
       share_name: the name of the share to backup
       namespace_name: namespace name of the remote object store service
       bucket_name: bucket name of the remote bucket on object store service
       obj_name: name of the backup object to be placed in the bucket

     Exception:
      raise StandardError if backup has failed
    """

    logger.debug('start of  create_share_backup for %s' % share_name)
    result = priv_exec.priv_exec('backup_share_dump',
                                 {'name': share_name,
                                  'namespace_name': namespace_name,
                                  'bucket_name': bucket_name,
                                  'obj_name': obj_name,
                                  'user_key': current_user.user_key})
    logger.debug('end of  create_share_backup for %s' % share_name)
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    return result


def restore_share_backup(share_name, namespace_name, bucket_name, obj_name):
    """
    Restore a backup of a share.
    A (XFS) dump is dowmloaded from remote object store and restored as
    the share content
    arguments :
       share_name: the name of the share to backup
       namespace_name: namespace name of the remote object store service
       bucket_name: bucket name of the remote bucket on object store service
       obj_name: name of the backup object to be placed in the bucket
    Exception:
      raise StandardError if restoration has failed
    """

    logger.debug('start of  restore_share_backup for %s' % share_name)
    result = priv_exec.priv_exec('restore_share_dump',
                                 {'name': share_name,
                                  'namespace_name': namespace_name,
                                  'bucket_name': bucket_name,
                                  'obj_name': obj_name,
                                  'user_key': current_user.user_key})
    logger.debug('end of  restore_share_backup for %s' % share_name)
    if result['results'] == 'failure':
        raise StandardError(result['reason'])
    logger.debug('restore_share_backup returning [%s]' % str(result))
    return result


def get_oci_config():
    """
    Gets OCI service access configuration for root user
    """
    result = priv_exec.priv_exec('get_oci_config',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def delete_oci_config():
    """
    Deletes OCI service access configuration(root/.oci) for root user
    Exception:
      raise StandardError if removal of /root/.oci has failed
    """
    result = priv_exec.priv_exec('delete_oci_config',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def get_background_task(task_id):
    """
    Returns ongoing task
    a task is
    {
       'ID' : unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status' : {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                  },
       'states' : [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
      }
    """
    result = priv_exec.priv_exec('get_task',
                                 {'ID': task_id,
                                  'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def list_background_tasks(**kargs):
    """
    Returns ongoing task list (can be empty)
    return tasks matching all specified parameters
    keyword parameter:
       'running_only':do we fetch only running tasks?
       'target_name': fetch only tasks for this target
       'task_name': fetch only tasks of that name
    a task is
    {
       'ID' : unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status' : {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                  },
       'states' : [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
      }
    """
    _params = {'user_key': current_user.user_key}
    if kargs.get('running') is not None:
        _params['running_only'] = kargs.get('running')
    if kargs.get('target_name') is not None:
        _params['target_name'] = kargs.get('target_name')
    if kargs.get('task_name') is not None:
        _params['task_name'] = kargs.get('task_name')

    result = priv_exec.priv_exec('get_task_list',
                                 _params)
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def set_oci_config(oci_user, oci_tenancy, oci_region, keyfile,
                   keyfile_passphrase):
    """
    Sets OCI service access configuration
    Parameters:
       oci_user: OCID of the user calling the API
       oci_tenancy: OCID of the tenancy.
       oci_region: An Oracle Cloud Infrastructure region.
       keyfile: private key file (optional) as FileStorage
       keyfile_passphrase: passphrase used on key file (optional)

    If private key file is provided, the one on local storage is overwritten
    raises StandardError: in case of error
    """

    # we consume the file here as StorageFile cannot be serialized
    # to be sent to priv_exec server

    # protect ourself against big file sent by user
    # we do not expect a key file to be more than 5k
    # we perform 1k read at a time as FileStorage.content_length is actually
    # not available
    keyfile_buffer = None
    keyfile_content = None
    if keyfile and keyfile.filename:
        keyfile_buffer = cStringIO.StringIO()
        keyfile_size = 0
        try:
            while True:
                data = keyfile.read(1024)
                if not data:
                    break
                keyfile_size += len(data)
                if keyfile_size > 5120:
                    raise StandardError('keyfile too big')
                keyfile_buffer.write(data)
                logger.debug('keyfile content size=%d, written to buffer' % keyfile_buffer.tell())
                keyfile_content = keyfile_buffer.getvalue()
        except Exception, e:
            logger.exception('Failed to read keyfile [%s]' % str(e))
            raise StandardError('Failed to read keyfile [%s]' % str(e))
        finally:
                if keyfile:
                    keyfile.close()
                if keyfile_buffer:
                    keyfile_buffer.close()

    result = priv_exec.priv_exec('set_oci_config',
                                 {
                                     'oci_user': oci_user,
                                     'oci_tenancy': oci_tenancy,
                                     'oci_region': oci_region,
                                     'oci_keyfile_content': keyfile_content,
                                     'oci_keyfile_passphrase': keyfile_passphrase,
                                     'user_key': current_user.user_key
                                 })
    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def get_oci_shape():
    '''
    Retrieve the OCI shape string associated with the running instance. For example: BM.DenseIO1.36, VM.DenseIO1.8 etc
    Output: shape as string
    Raise: StandardError in case of failure
    '''
    instance = OCIInstanceHelper()
    return OCIInstanceHelper.get_oci_shape(instance)


def get_platform_status():
    """
    Checks OCI shape to see if the platform is supported by LSA
    raises:
      StandardError: unsupported platform <OCI shape>
    """
    return {'results': 'success'}
    # TODO cleanup, need to clean up front end calls and management codes
    shape_list = ['BM.HighIO1.36',
                  'BM.DenseIO1.36',
                  'VM.DenseIO1.4',
                  'VM.DenseIO1.8',
                  'VM.DenseIO1.16',
                  'BM.DenseIO2.52',
                  'VM.DenseIO2.8',
                  'VM.DenseIO2.16',
                  'VM.DenseIO2.24']
    instance = OCIInstanceHelper()
    try:
        shape = OCIInstanceHelper.get_oci_shape(instance)
        if shape in shape_list:
            result = {'results': 'success'}
        else:
            result = {'results': 'failure', 'reason': 'unsupported platform ' + shape}
    except StandardError:
        # not in OCI, but continue for testing
        result = {'results': 'success'}

    if result['results'] == 'failure':
        raise StandardError(result['reason'])

    return result


def unconfigure():
    result = priv_exec.priv_exec('unconfigure_platform',
                                 {'user_key': current_user.user_key})
    if result['results'] == 'failure':
        raise StandardError(result['reason'])


# tests
if __name__ == '__main__':
    # import pdb; pdb.set_trace()

    # in case we run as standalone app
    # init logging
    try:
        set_log_config()
    except Exception, e:
        print 'cannot setup logging: %s' % str(e)
    # print get_nfs_clients()2
    # print getServices()
