"""
module hosting login blueprint views
see https://confluence.oraclecorp.com/confluence/display/BSA/custom+block+storage+support
This blueprint handle /login and /logout URLs (views)
"""
__copyright__ = "Copyright 2017, 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from flask import Blueprint, render_template, request, redirect, url_for

from flask.ext.login import login_required, login_user, logout_user, current_user

from urlparse import urlparse, urljoin

import priv_exec_server as priv_exec

from flask import current_app

from ..doc.views import auto


from ..security import unregister_user, register_user, is_admin_created, User

login = Blueprint('login', __name__,
                  template_folder='templates',
                  static_folder='static')


@login.route('/', methods=['GET', 'POST'])
@auto.doc(groups=['internal'])
def loginuser():
    ''' returns login screen on GET and perform authentication on POST '''

    if not is_admin_created():
        return redirect(url_for('setup.create_admin'))

    if current_user.is_authenticated:
        # already authenticated send user back to index
        return redirect(url_for('bootstrap.index'))

    current_app.logger.debug('inside loginuser for method [%s]' % request.method)
    if request.method == 'GET':
        return render_template('alta_login.html', error_message="",
                               lsa_product_name="Oracle Linux Storage Appliance")

    if 'username' not in request.form:
        raise KeyError('Missing username parameter')
    if 'password' not in request.form:
        raise KeyError('Missing password parameter')

    res = priv_exec.priv_exec('authenticate',
                              {'username': request.form['username'],
                               'password': request.form['password'],
                               'user_key': current_user.user_key})
    current_app.logger.debug('result %s' % res['results'])

    if res['results'] == 'success':
        u = User(request.form['username'], res['user_key'])
        register_user(u)
        login_user(u)
    else:
        current_app.logger.warning(res['reason'])
        return render_template('alta_login.html', lsa_product_name="Oracle Linux Storage Appliance", error_message=res['reason'])

    request_url = get_redirect_target()

    if request_url:
        current_app.logger.debug('redirect to request_url=[%s]' % request_url)
        return redirect(request_url)
    current_app.logger.debug('redirect to index=[%s]' % url_for('bootstrap.index'))
    return redirect(url_for('bootstrap.index'))


@login.route("/logout")
@auto.doc(groups=['internal'])
@login_required
def logout():
    ''' log current user out returns the URL caller should move to'''
    current_app.logger.info('login out')
    priv_exec.priv_exec('logout',
                        {'username': current_user.id,
                         'user_key': current_user.user_key})
    unregister_user(current_user)
    logout_user()

    # should be a call to 'redirect(url_for('bootstrap.index'))'
    # but ajax calls do not honor redirect: redirect not triggered on client side
    # return the new URL the caller should move to instead
    return url_for('bootstrap.index')


def is_safe_url(target):
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and ref_url.netloc == test_url.netloc


def get_redirect_target():
    ''' retrieve the previous URL if present '''
    whitelist = ['login', 'logout']
    for target in request.values.get('next'), request.referrer, request.url:
        if not target:
            continue
        if any(x in target for x in whitelist):
            continue
        if is_safe_url(target):
            return target
