/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore', 'knockout', 'jquery',
        'models/dashboard', 'models/storagepool','viewModels/common_vm_def',
        'models/platform', 'models/nfs','ojs/ojcollapsible',
        'ojs/ojgauge', 'ojs/ojtable', 'ojs/ojarraytabledatasource'],
       function(oj, ko, $,dashboard, storagePool, CommonVMDefinitons,
                Platform, NFS) {

    function DashboardViewModel() {
      var self = this;


      self.top = new ko.observableArray([]);
      self.topTable = new oj.ArrayTableDataSource(self.top, {idAttribute: 'name'});

      self.cpu = ko.observable(0);
      self.mem = ko.observable(0);
      self.disk = ko.observable(0);
      self.net_sent = 0;
      self.net_recv = 0;
      self.net_sent_prev = 0;
      self.net_recv_prev = 0;
      self.bytes_sent = ko.observable();
      self.bytes_recv = ko.observable();
      self.clientConnected = ko.observable('Retrieving number of NFS clients connected');

     self.cpuMessage = ko.computed(function() {
         return (self.cpu() + "%");
      });


      this.poolSize = ko.observable(0);
      this.poolFreeSize = ko.observable(0);

      /* returns the percentage of pool storage consumed */

      this.poolConsumed = ko.pureComputed(function() {
          if (self.poolSize() > 0 && self.poolFreeSize() > 0) {
              return (((self.poolSize()-self.poolFreeSize())*100)/self.poolSize());
          }
          return 0.0;
      });

      function updatePoolUsage() {
        storagePool.getPoolUsage().then (function (data) {
          if ('error' in data || (total === 0 || free === 0)) {
            self.poolSize(0);
            self.poolFreeSize(0);
            self.storageUsageTitle('error while retrieving pool usage');
            return;
          }

          var total = data.total;
          var free = data.free;
          self.poolSize(total);
          self.poolFreeSize(free);
          var formattedFree = self.byteConverter.format(free);
          var formattedTotal = self.byteConverter.format(total);
          self.storageUsageTitle("Storage utilization: " + formattedFree +
                                 ' free out of ' + formattedTotal);
        }).catch(function(raeson) {
          oj.Logger.warn('get pool usage has failed : %s',raeson);
        });
      }
      updatePoolUsage();

      self.storageUsageTitle = ko.observable();
      self.cpuUsage = ko.observable('CPU usage: retrieving information');
      self.cpuCount = ko.observable('unknown');
      self.memTotal = ko.observable('unknown');
      self.memAvail = ko.observable(0);
      self.diskTotal = ko.observable('unknown');
      self.diskAvail = ko.observable(0);
      self.stop = false;

      var p = new Platform();
      p.getInfos().then(function(info) {
        self.cpuCount(info.cpu);
        self.memTotal(info.mem);
        self.diskTotal(info.root_disk_size);
      });

      this.cpuUsageTitle = ko.pureComputed(function () {
            return 'CPU usage: ' + self.cpu() + '% out of ' + self.cpuCount() + ' CPUs';
      });

      this.memUsageTitle = ko.pureComputed(function () {
          var fmtMemAvail = self.byteConverter.format(self.memAvail());
          var fmtMemAll = self.byteConverter.format(self.memTotal());
          return 'Memory usage: ' + fmtMemAvail + ' available out of ' + fmtMemAll;
      });

      this.diskUsageTitle = ko.pureComputed(function() {
          var fmtDiskAvail = self.byteConverter.format(self.diskAvail());
          var fmtDiskAll = self.byteConverter.format(self.diskTotal());
          return 'Appliance only disk usage: ' + fmtDiskAvail + ' free out of ' + fmtDiskAll;
      });

      function usagesRefreshFunc() {
            new Platform().getAllUsages().then(function (data) {
                self.cpu(data.cpu);
                self.mem(data.mem);
                self.memAvail(data.mem_avail);
                self.disk(data.disk);
                self.diskAvail(data.disk_avail);
                self.net_sent_prev = self.net_sent;
                self.net_recv_prev = self.net_recv;

                self.net_sent = data.bytes_sent;
                self.net_recv = data.bytes_recv;

                if (self.net_sent_prev != 0 || self.net_recv_prev != 0) {
                    var sent_mb_s = ((self.net_sent - self.net_sent_prev)/self.UI_UPDATE_BASE_TIME)/1e6;
                    var recv_mb_s = ((self.net_recv - self.net_recv_prev)/self.UI_UPDATE_BASE_TIME)/1e6;
                    self.bytes_sent(sent_mb_s.toFixed(2));
                    self.bytes_recv(recv_mb_s.toFixed(2));
                }
            });
        }
        var usagesRefresh = undefined;
        function nfsClientsRefreshFunc() {
            NFS.getClientCount().then(function (count) {
                if (count !== 0) {
                    self.clientConnected(count + ' NFS client' +
                                         (count > 1 ? 's' : '') + ' connected');
                }else{
                    self.clientConnected('No NFS client connected');
                }
            });
            updatePoolUsage();
        }
        var nfsClientsRefresh = undefined;


      self.handleDeactivated = function () {
          window.clearInterval(nfsClientsRefresh);
          window.clearInterval(usagesRefresh);
      };
      self.handleActivated = function(info) {
      storagePool.getShares().then(function (data) {
            self.tmpTop = data.map(function(share) {
                return {'name': share.name,
                        'size': share.total,
                        'free': share.free,
                        'title': self.byteConverter.format(share.spaceAvailable) + ' free out of ' + self.byteConverter.format(share.diskSize),
                        'util': share.spaceAvailable / share.diskSize,
                        'type': 'NFS'
                       };
            });
        });
        nfsClientsRefresh = window.setInterval(nfsClientsRefreshFunc,
                                                 self.UI_UPDATE_BASE_TIME * 2000);
        usagesRefresh = window.setInterval(usagesRefreshFunc,
                                             self.UI_UPDATE_BASE_TIME * 1000);
      };
    }

    $.extend(DashboardViewModel.prototype,CommonVMDefinitons.prototype.definitions);
    $.extend(DashboardViewModel.prototype,CommonVMDefinitons.prototype.converters);

    DashboardViewModel.prototype.gaugeCPULabel =
            oj.Translations.getTranslatedString('CPU');
    DashboardViewModel.prototype.gaugeMemoryLabel =
            oj.Translations.getTranslatedString('Memory');
    DashboardViewModel.prototype.gaugeDiskLabel =
            oj.Translations.getTranslatedString('Disk');

    DashboardViewModel.prototype.netUnitLabel =
            oj.Translations.getTranslatedString('MB/s');


    DashboardViewModel.prototype.netInputLabel =
            oj.Translations.getTranslatedString('Network Input');
    DashboardViewModel.prototype.netOutputLabel =
            oj.Translations.getTranslatedString('Network Output');

    DashboardViewModel.prototype.storageStatusSectionLabel =
               oj.Translations.getTranslatedString('Storage Status');
    DashboardViewModel.prototype.applianceStatusSectionLabel =
               oj.Translations.getTranslatedString('Appliance Status');


    return new DashboardViewModel();
  }
);
