#!/usr/bin/python
__copyright__ = "Copyright 2017, 2018, Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
import asyncore
import socket
import json
import pexpect
import uuid
import pam
import getpass

import os
import re
import subprocess
import StringIO

import Queue

import struct
import shutil
import time

import oci.config
import oci.object_storage

from oci_utils import iscsiadm, lsblk
import oci_utils
from oci.regions import REGIONS_SHORT_NAMES as oci_regions
from backend.utility import format_readable_to_bytes

from backend.nfs_util import restart_nfs_services, NfsExportEntry
from backend.service_util import SystemdService
import backend.ssh_util
from backend.lvm_util import (get_volume_group_stats,
                              setup_snapshot,
                              remove_snapshot,
                              list_vg_disks,
                              get_physical_volumes
                              )
import backend.samba_util
from backend.xfs_util import (get_restore_command,
                              get_dump_command,
                              get_resume_restore_command,
                              RestoreStreamExecutor,
                              DumpStreamExecutor)
import backend.taskmgr
from backend.taskmgr.task_manager import TaskManager
from backend.taskmgr import (TaskStatusCode, TaskStatus)
from backend.object_store_util import (StreamMultiPartUploader, StreamDownloader)
import backend.system_util
import logging
from backend.lsa_types import StreamQueue
from backend.common import DEV_PATH_BY_ID
from backend.common import (SYSCTL, PRIV_EXEC_LOG_FILENAME, SAFE_SOCKET_OP_TIMEOUT, DEFAULT_LOG_LINES,
                            CMD_RPM, SHARE_POINT, STREAM_QUEUE_LENGTH, VAR_DIR, TASK_MGR_DB_FILE, RAID_DEV,
                            RAID_LEVEL_NONE, SMB_EXPORT_CONF)
from backend.common import (CMD_USER_ADD, CMD_USER_DEL, CMD_PASSWD, CMD_OPENSSL)
from backend.common import (LSA_STATE_BIN, LSA_STATE_UNCONFIGURED)

from backend.log_util import set_log_config
from backend.oci_config_util import OCIConfigParser

from backend import (BackupTask, RestoreTask, ExecuteAfterThread)
from backend.lsa_types import ExtJSONEncoder as LSAExtJSONEncoder
import backend.share_util
from backend.config_manager import NfsConfigManager, NFSManager, SambaConfigManager, SambaManager
from backend.samba_util import SambaExportEntry
import backend.api_wrapper
from backend.storage_pool import StoragePool
from backend.share_util import ShareEntry
from backend.lvm_util import is_vg_created
from backend.command_util import execute_command

logger = logging.getLogger('lsa.privserver')


# how to add a new function
# in cmd_mapping at a key pair with the function unique id as a string
# and a reference to the function to be called
# the function should return at least:
# on success: {'results': 'success'}
# see  RequestGoodResult
# on failure: {'results': 'failure', 'reason': 'meaningful reason'}
# see RequestBadResult

# list of current authenticated users key
# anonymous always present
current_user_keys = ['anonymous']


def check_credentials(username, password):
    ''' return true is username and password are a valid combination '''
    ''' use pam if running as root otherwise su test'''
    if getpass.getuser() == 'root':
        return pam.authenticate(username, password, service='system-auth')
    else:
        return check_credentials_no_root(username, password)


def check_credentials_no_root(username, password):
    ''' return true is username and password are a valid combination '''
    # avoid cmd overflow
    logger.debug('su auth')
    if ' ' in username:
        return False
    result = -1

    try:
        child = pexpect.spawn('/bin/su - %s -c \"echo \'blah\'\"' % (username))
        child.expect('Password:')
        child.sendline(password)
        result = child.expect(['blah', 'su: Authentication failure'], timeout=5)
        child.close()
    except:
        child.close()

    if result == 0:
        return True
    return False


def authenticate(params):
    if not all(p in params for p in ("username", "password")):
        return RequestBadResult('missing parameters').toDict()
    username = params['username']
    password = params['password']
    if check_credentials(username, password):
        user_key = str(uuid.uuid4())
        current_user_keys.append(user_key)
        return RequestGoodResult({'user_key': user_key}).toDict()
    else:
        logger.warning('authentication failed due to wrong credentials for user [%s]' % username)
        return RequestBadResult('invalid credentials').toDict()


def change_password(params):
    """
    Change the system password this the logged in user.
    expected arguments:
        username: username
        old_password: previous password
        new_password: new password
    """
    username = params['username']
    old_password = params['old_password']
    new_password = params['new_password']
    if not check_credentials(username, old_password):
        return RequestBadResult('invalid credentials, old password mismatched').toDict()

    (out, err, res) = execute_command([CMD_PASSWD, '--stdin', username], input=new_password)
    if res == 0:
        return RequestGoodResult().toDict()
    else:
        logger.error('passwd command exited on %d: %s' % (res, err))
        return RequestBadResult('update of user password has failed: %s' % err).toDict()


def logout(params):
    if not all(p in params for p in ("username", "user_key")):
        return {'results': 'failure', 'reason': 'missing parameters'}
    key = params['user_key']
    username = params['username']
    if key in current_user_keys:
        current_user_keys.remove(key)
        return RequestGoodResult({'username': username}).toDict()
    else:
        return RequestBadResult('user_key not present').toDict()


def test_priv(params):
    ''' example function create a file with root creds'''
    if not all(p in params for p in ("filename", "user_key")):
        return RequestBadResult('missing parameters').toDict()
    logger.debug('params %s' % params.get('filename'))
    subprocess.call(['/usr/bin/rm', '/tmp/test_priv'])
    subprocess.call(['/bin/touch', '/tmp/test_priv'])
    return RequestGoodResult({'blah': 'test_priv_created'}).toDict()


def create_pool(params):
    '''
    Create the pool in backend storage
    expected parameters:
        raid_level: raid level (0, 1, 6, 10, or RAID_LEVEL_NONE)
        devices: list of comma-separated paths (e.g., '/dev/sdb,/dev/sdc')
    '''
    raid_level = params.get('raid_level')
    if raid_level not in [0, 1, 6, 10, RAID_LEVEL_NONE]:
        return RequestBadResult('unsupported parameter raid_level').toDict()
    devices = params.get('devices')
    if not devices:
        return RequestBadResult('missing parameter devices').toDict()
    try:
        backend.lvm_util.create_pool(raid_level, devices)
        return RequestGoodResult().toDict()
    except Exception, e:
        logger.debug('Creating pool failed: %s' % str(e))
        return RequestBadResult('Creating pool failed: %s' % str(e)).toDict()


def list_shares(params):
    """
    list the shares
    expected parameters:
        names: list of names  of the share we want to retrieve
    """
    shares = None
    names = params.get('names')
    try:
        shares = backend.share_util.get_shares(names)
    except Exception, e:
        logger.exception('cannot list shares: %s' % str(e))
        return RequestBadResult('cannot list shares: %s' % str(e)).toDict()
    return RequestGoodResult({'shares': [share.serialize() for share in shares]}).toDict()


def create_share(params):
    '''
    Create a share in backend storage
      parameter:
        name: name of the share
        size: Size of the share in bytes
        inode32: workaround for legacy application limitation for 26879349
    Resolve on newly created share's name
      returns: dictionary as {'task_id':<task id>}
    '''
    share_name = params.get('name')
    if not share_name:
        return RequestBadResult('missing name parameter').toDict()
    size = params.get('size')
    if not size:
        return RequestBadResult('missing size parameter').toDict()
    inode32 = params.get('inode32')
    if inode32 is None:
        return RequestBadResult('missing inode32 parameter').toDict()
    try:
        task = backend.share_util.AsyncShareCreateTask(share_name, size, inode32)
        logger.debug('after share_util.AsyncShareCreateTask()')
        backend.taskmgr.get_task_manager().register_task(task)
        task.begin()
        logger.debug('after task.begin(): %s' % str(task.get_id()))
        return RequestGoodResult({'task_id': task.get_id()}).toDict()
    except Exception, e:
        logger.debug('Adding share failed: %s' % str(e))
        return RequestBadResult('Adding share failed: %s' % str(e)).toDict()


def resize_share(params):
    '''
    Modify a share in backend storage
      parameter:
        name: name of the share
        size: Size of the share in bytes
    '''
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()
    size = params.get('size')
    if not size:
        return RequestBadResult('missing size parameter').toDict()
    try:
        backend.share_util.resize_share(name, size)
        return RequestGoodResult().toDict()
    except Exception, e:
        logger.debug('Modifying share failed: %s' % str(e))
        return RequestBadResult('Modifying share failed: %s' % str(e)).toDict()


def delete_share(params):
    '''
    Delete a share. We delete the associated exports first before deleting the share
        parameter:
            name: name of the share
    '''
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()
    try:
        share = ShareEntry(name)
        nfs_conf_manager = NfsConfigManager()
        smb_conf_manager = SambaConfigManager(SMB_EXPORT_CONF)
        smb_manager = SambaManager()
        nfs_manager = NFSManager()
    except Exception, e:
        logger.exception(str(e))
        return RequestBadResult(str(e)).toDict()
    try:
        exports = nfs_conf_manager.get_exports(share)
        # delete all the NFS exports associated with the share
        nfs_conf_manager.unexport_share(share)
        # remove entries from etab
        nfs_manager.remove_export_list(share, exports)
        # Delete the samba exports
        smb_conf_manager.remove_exports_by_share(share)
        # check samba services after config changes
        code = smb_manager.remove_export_list(share, exports)
        if code != 0:
            smb_conf_manager.revert_smb_conf_modifications()
            smb_manager.remove_export_list(share, exports)
        # Now delete the LV, FS directory etc
        backend.share_util.delete_share(name)
    except Exception, e:
        logger.exception('Deleting share failed: %s' % str(e))
        return RequestBadResult('Deleting share failed: %s' % str(e)).toDict()
    return RequestGoodResult().toDict()


def delete_nfs_export(params):
    """
    Delete an NFS export on the system
    expected argument:
        'name': The name of the share
        'hosts': hosts the share is exported to
    """
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()
    hosts = params.get('hosts')
    if not hosts:
        return RequestBadResult('missing name parameter').toDict()
    logger.info('delete_nfs_export [%s]' % name)
    try:
        share = ShareEntry(name)
        nfs_manager = NfsConfigManager()
        export = nfs_manager.get_export(share, hosts)
        if export is not None:
            nfs_manager.remove_export(share, export)
            NFSManager().remove_export(share, export)
    except Exception, e:
        logger.debug('remove of NFS share has failed: %s' % str(e))
        return RequestBadResult('remove of NFS share has failed: %s' % str(e)).toDict()

    return RequestGoodResult().toDict()


def create_nfs_export(params):
    """
    Adds an NFS export on the system
    expected argument:
        'name': The name for the new share
        'hosts': authorized clients for the share
        'options': the share options as expected by NFS( dictionary item
                            example {sync: true, hide: false, root_squash: true} )

    """
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()

    hosts = params.get('hosts')
    if not hosts:
        return RequestBadResult('missing hosts parameter').toDict()
    options = params.get('options')
    if not options:
        return RequestBadResult('missing options parameter').toDict()

    logger.debug('add_nfs_export [%s/%s/%s]' % (name, hosts, options))
    try:
        share = ShareEntry(name)
        export = NfsExportEntry(hosts, share.get_path(), **options)
        nfs_conf_manager = NfsConfigManager()
    except Exception, e:
        return RequestBadResult(e.args[0]).toDict()
    try:
        nfs_conf_manager.add_export(share, export)
    except Exception, e:
        logger.exception('add export on NFS config Manager failed')
        return RequestBadResult(e.args[0]).toDict()
    try:
        NFSManager().add_export(share, export)
    except (ValueError, StandardError) as e:
        logger.exception('add export on NFS Manager failed')
        # cleanup the export that we added
        nfs_conf_manager.remove_export(share, export)
        return RequestBadResult(e.args[0]).toDict()
    return RequestGoodResult({'entry': export.to_json()}).toDict()


def create_samba_export(params):
    '''
    Add a samba export for a particular share
    expected arguments:
        'name': name of the share
        'resource': resource name of the samba export
        'options': samba export options as dictionary items. Format example
            {"comment": "comment", "browsable": "boolean", "guest_ok": "boolean", "hosts_allow": "x.x.x.x",
             "strict_sync": "", "sync_always": "boolean", "hosts_deny": "x.x.x.x", "read only": "boolean"}
    output: SambaExportEntry object as dictionary
    '''
    logger.debug('samba export creation [%s]' % params)
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()
    resource = params.get('resource')
    if not resource:
        return RequestBadResult('missing resource parameter').toDict()
    options = params.get('options')
    if not options:
        return RequestBadResult('missing options parameter').toDict()
    try:
        smb_conf_manager = SambaConfigManager(SMB_EXPORT_CONF)
        smb_manager = SambaManager()
        share = ShareEntry(name)
        exp = SambaExportEntry(resource, **options)
    except Exception, e:
        logger.exception('Failed to create samba export: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()

    try:
        smb_conf_manager.add_export(share, exp)
    except Exception as e:
        logger.exception('create_samba_export failed')
        return RequestBadResult(str(e)).toDict()

    try:
        smb_manager.add_export(share, exp)
    except Exception as e:
        logger.exception('Failed to add samba selinux rules')
        return RequestBadResult(str(e)).toDict()
    # try to see if we need to remove the export by checking the services after config file modification
    code = smb_manager.remove_export(share, exp)
    if code != 0:
        smb_conf_manager.revert_smb_conf_modifications()
        smb_manager.remove_export(share, exp)
        return RequestBadResult('Creating samba export failed. Config file modifications error, failed to restarted services, kept the original config file').toDict()

    return RequestGoodResult({'entry': exp.to_json()}).toDict()


def modify_samba_export(params):
    '''
    Modify a samba export for a particular share
    expected arguments:
        'name': name of the share
        'resource': resource name of the samba export
        'options': samba export options as dictionary items. Format example
            {"comment": "comment", "browsable": "boolean", "guest_ok": "boolean", "hosts_allow": "x.x.x.x",
             "strict_sync": "", "sync_always": "boolean", "hosts_deny": "x.x.x.x", "read_only": "boolean"}
    output: SambaExportEntry object as dictionary
    '''
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()

    resource = params.get('resource')
    if not resource:
        return RequestBadResult('missing resource parameter').toDict()
    options = params.get('options')
    if not options:
        return RequestBadResult('missing options parameter').toDict()

    try:
        smb_conf_manager = SambaConfigManager(SMB_EXPORT_CONF)
        smb_manager = SambaManager()
        share = ShareEntry(name)
        exp = SambaExportEntry(resource, **options)
    except Exception, e:
        logger.exception('Failed to modify samba export: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()
    try:
        smb_conf_manager.modify_export(share, exp)
    except Exception as e:
        logger.exception('modify_samba_export failed')
        return RequestBadResult(str(e)).toDict()
    # try to see if we need to remove the export by checking the services after config file modification
    code = smb_manager.remove_export(share, exp)
    if code != 0:
        smb_conf_manager.revert_smb_conf_modifications()
        smb_manager.remove_export(share, exp)
        return RequestBadResult('Modifying samba export failed. Config file modifications error, failed to restarted services, kept the original config file').toDict()

    return RequestGoodResult({'entry': exp.to_json()}).toDict()


def delete_samba_export(params):
    '''
    Delete a samba export
    expected arguments:
        'resource': resource name of the samba export(Identifier)
    '''
    resource = params.get('resource')
    if not resource:
        return RequestBadResult('missing resource parameter').toDict()
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()
    try:
        share = ShareEntry(name)
        smb_conf_manager = SambaConfigManager(SMB_EXPORT_CONF)
        smb_manager = SambaManager()
    except Exception, e:
        logger.exception('Failed to delete samba export: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()
    try:
        exp = smb_conf_manager.get_export(share, resource)
        if exp is None:
            return RequestBadResult('Samba Export not foudn in the cpnfig file').toDict()
    except Exception, e:
        return RequestBadResult(e.args[0]).toDict()

    try:
        smb_conf_manager.remove_export(exp)
    except Exception as e:
        logger.exception('delete_samba_export failed')
        return RequestBadResult(str(e)).toDict()

    code = smb_manager.remove_export(share, exp)
    if code != 0:
        smb_conf_manager.revert_smb_conf_modifications()
        smb_manager.remove_export(share, exp)
        return RequestBadResult('Deleting samba export failed. Config file modifications error, failed to restarted services, kept the original config file').toDict()

    return RequestGoodResult().toDict()


def backup_share_dump(params):
    """
    Dumps a share and backup it on remote object store
    expected arguments:
       'name': The name for the share
       'namespace_name': namespace name to be used on remote object store
       'bucket_name': name of the bucket to be used on remote object store
       'objName':  name of the backup object to placed in the bucket
    Operations happen in background
    """
    share_name = params.get('name')
    if not share_name:
        return RequestBadResult('missing name parameter').toDict()
    namespace_name = params.get('namespace_name')
    if not namespace_name:
        return RequestBadResult('missing namespace_name parameter').toDict()
    bucket_name = params.get('bucket_name')
    if not bucket_name:
        return RequestBadResult('missing bucket_name parameter').toDict()
    object_name = params.get('obj_name')
    if not object_name:
        return RequestBadResult('missing objName parameter').toDict()

    logger.debug('share_name=%s, namespace_name=%s,bucket_name=%s, object_name=%s' %
                 (share_name, namespace_name, bucket_name, object_name))

    try:
        client = oci_utils.oci_api.OCISession().get_object_storage_client()
    except Exception, e:
        logger.debug('failed to access object storage: %s' % str(e))
        return RequestBadResult('failed to access object storage: %s' % str(e)).toDict()

    share_snapshot = ''
    try:
        share_snapshot = setup_snapshot(share_name)
    except Exception, e:
        logger.exception('cannot create snapshot')
        return RequestBadResult(e.args[0]).toDict()

    bck_task = BackupTask(share_name)
    backend.taskmgr.get_task_manager().register_task(bck_task)

    logger.debug('new task created for backup %d' % bck_task.get_id())

    # creates and start the XFS dumper
    dumper = get_dump_command(os.path.join(SHARE_POINT, share_snapshot))
    logger.debug('new dumper created %s' % str(dumper))
    queue = StreamQueue(STREAM_QUEUE_LENGTH)
    dumper_executor = DumpStreamExecutor(dumper, queue)
    logger.debug('new dumper executor created')

    try:
        dumper.start()
        logger.debug('dumper started')
        dumper_executor.start()
        logger.debug('dump executor started')

        # creates and start uploaders
        uploader = StreamMultiPartUploader(bck_task, queue, client, namespace_name, bucket_name, object_name)
        # remove share_snapshot when the uploader thread is finished
        end_th = ExecuteAfterThread(uploader, remove_snapshot, (share_snapshot,))
        uploader.init_upload()
        uploader.start()
        end_th.start()
        logger.debug('uploader started')
    except StandardError, e:
        logger.exception('cannot start uploader')
        dumper_executor.stop()
        dumper.cleanup(True)
        bck_task.end(backend.taskmgr.TaskStatus(backend.taskmgr.TaskStatusCode.FAILED,
                                                'backup failed: %s' % str(e)))
        remove_snapshot(share_snapshot)
        return RequestBadResult(e.args[0]).toDict()

    return RequestGoodResult({'taskId': bck_task.get_id()}).toDict()


def restore_share_dump(params):
    """
    restore previously created dump of a share.
    Operations are done asynchronously.
    expected arguments:
       'name': The name for the share
       'namespace_name': the name of namespace used on remote object storage service
       'bucket_name': name of the bucket to be used on remote object store
       'objName':  name of the backup object to placed in the bucket
    returns the name of the path of the dump file
     as {'dumpPath': <path>}
    """
    share_name = params.get('name')
    if not share_name:
        return RequestBadResult('missing name parameter').toDict()
    namespace_name = params.get('namespace_name')
    if not namespace_name:
        return RequestBadResult('missing namespace_name parameter').toDict()
    bucket_name = params.get('bucket_name')
    if not bucket_name:
        return RequestBadResult('missing bucket_name parameter').toDict()
    object_name = params.get('obj_name')
    if not object_name:
        return RequestBadResult('missing obj_name parameter').toDict()

    try:
        client = oci_utils.oci_api.OCISession().get_object_storage_client()
    except Exception, e:
        logger.debug('failed to access object storage: %s' % str(e))
        return RequestBadResult('failed to access object storage: %s' % str(e)).toDict()

    restore_task = RestoreTask(share_name)
    backend.taskmgr.get_task_manager().register_task(restore_task)
    logger.debug('new task created for restore %d' % restore_task.get_id())

    # first be sure no previous interrupted restore will stop us
    # this command wait on stdin, by closing it we end the command
    resumer = get_resume_restore_command(os.path.join(SHARE_POINT, share_name))
    resumer.start()
    resumer.get_input_stream().close()
    resumer.cleanup()
    logger.debug('resumed any left restore')

    # creates and start the XFS restorer
    restorer = get_restore_command(os.path.join(SHARE_POINT, share_name))
    logger.debug('new restorer created %s' % str(restorer))
    queue = StreamQueue(STREAM_QUEUE_LENGTH)
    restorerExecutor = RestoreStreamExecutor(restorer, restore_task, queue)

    restorer.start()
    logger.debug('restorer started')
    restorerExecutor.start()
    logger.debug('restore executor started')

    # creates and start downloaderS
    try:
        downloader = StreamDownloader(queue, client, namespace_name,
                                      bucket_name, object_name)
        downloader.init_download()
        downloader.start()
        logger.debug('downloader started')
    except StandardError, e:
        logger.exception('cannot start downloader')
        restorerExecutor.stop()
        restorer.cleanup(True)
        restore_task.end(TaskStatus(TaskStatusCode.FAILED,
                                    'cannot start downloader: %s' % str(e)))
        return RequestBadResult(e.args[0]).toDict()

    return RequestGoodResult({'taskId': restore_task.get_id()}).toDict()


def modify_nfs_export(params):
    """
    modify a NFS share on the system
    expected argument:
        'name': The name for the share
        'hosts': current authorized client for the nfs export
        'options': the share options as expected by NFS
    """
    name = params.get('name')
    if not name:
        return RequestBadResult('missing name parameter').toDict()

    hosts = params.get('hosts')
    if not hosts:
        return RequestBadResult('missing hosts parameter').toDict()

    options = params.get('options')
    if not options:
        return RequestBadResult('missing options parameter').toDict()

    # convert options to dictionary
    logger.debug('modify_nfs_share [%s/%s/%s]' % (name, hosts, options))

    try:
        share = ShareEntry(name)
        export = NfsExportEntry(hosts, share.get_path(), **options)
        nfs_conf_manager = NfsConfigManager()
        nfs_manager = NFSManager()
    except Exception, e:
        logger.exception('modifying nfs export failed: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()

    try:
        nfs_conf_manager.update_export(share, export)
    except Exception, e:
        logger.exception('modifying nfs export failed: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()
    try:
        nfs_manager.remove_export(share, export)
        nfs_manager.add_export(share, export)
    except (ValueError, Exception) as e:
        logger.exception('modify_nfs_export failed: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()
    return RequestGoodResult({'entry': export.to_json()}).toDict()


def compute_fingerprint(passphrase=None):
    '''
    Compute the fingerprint of the oci public key.
    The private key file is OCIConfigParser.LOCAL_KEY_FILE.
    Two steps are required to compute the fingerprint of the public key
    1. get the public from the private key
        #openssl rsa -pubout -outform DER -in private_key_file [-passin pass:passphrase]
    2. compute the fingerprint
        #openssl md5 -c public_key //public_key is output of step 1
    If computation of the fingerprint failed, set_oci_config() will return failure to the user.
    Note: this computation only ensures fingerprint can be computed out, it doesn't ensure the
    value of the fingerprint is correct. The validation will be done in separate step in set_oci_config()
    to ensure the correctness of fingerprint
    input:
        passphrase to protect the private key. It is optional
    output:
        None if calls to the above 2 cmds fail
        fingerprint if everything works fine. eg. 22:69:7a:01:b3:12:ab:21:7e:66:aa:e2:af:4f:6d:e3
    '''
    argcmd = [CMD_OPENSSL, 'rsa', '-pubout', '-outform', 'DER', '-in', OCIConfigParser.LOCAL_KEY_FILE]
    if passphrase:
        argcmd.extend(['-passin', 'pass:' + passphrase])
    (out, err, res) = execute_command(argcmd)
    if res > 0:
        logger.error('failed to get public key from private key [%s]' % err)
        return None
    (out, err, res) = execute_command([CMD_OPENSSL, 'md5', '-c'], input=out)
    if res > 0:
        logger.error('failed to get fingerprint of the public key [%s]' % err)
        return None

    fingerprint = out.split('=')[1].strip()
    logger.debug('fingerprint(compute) %s' % fingerprint)
    return fingerprint


def set_oci_config(params):
    """
    Sets OCI service access configuration
    Parameters:
       oci_user: OCID of the user calling the API
       oci_tenancy: OCID of the tenancy.
       oci_region: An Oracle Cloud Infrastructure region.
       oci_keyfile_content: private key file (optional) content as string
       oci_keyfile_passphrase: passphrase used on key file (optional)

    filename will be copied to local user key file
    If private key file is provided, the one on local storage is overwritten
    raises StandardError: in case of error
    """

    logger.debug('setting oci configuration')
    oci_user = params.get('oci_user')
    if not oci_user:
        return RequestBadResult('missing oci_user parameter').toDict()
    oci_tenancy = params.get('oci_tenancy')
    if not oci_tenancy:
        return RequestBadResult('missing  oci_tenancy parameter').toDict()
    oci_region = params.get('oci_region')
    if not oci_region:
        return RequestBadResult('missing oci_region parameter').toDict()
    oci_keyfile = params.get('oci_keyfile_content')
    oci_keyfile_passphrase = params.get('oci_keyfile_passphrase')

    # what is not specified as inputs we keep defaults
    # i.e what in the configuration already
    # first load what's exists
    config = OCIConfigParser()
    if os.path.exists(OCIConfigParser.CONFIG_FILE):
        config.load()

    config.set_option(OCIConfigParser.USER_OPT, oci_user)
    config.set_option(OCIConfigParser.TENANCY_OPT, oci_tenancy)
    config.set_option(OCIConfigParser.REGION_OPT, oci_region)

    if not os.path.exists(OCIConfigParser.LOCAL_CONF_HOME):
        logger.debug('configuration directory missing, create it')
        os.makedirs(OCIConfigParser.LOCAL_CONF_HOME, 0600)

    _keyfile_backup = OCIConfigParser.LOCAL_KEY_FILE + '.bck'

    # if we have key file provided,overwrite the one we have
    if oci_keyfile:
        # take care of passphrase of any
        if oci_keyfile_passphrase:
            config.set_option(OCIConfigParser.KEY_FILE_PASS_OPT, oci_keyfile_passphrase)

        if os.path.exists(OCIConfigParser.LOCAL_KEY_FILE):
            shutil.copyfile(OCIConfigParser.LOCAL_KEY_FILE, _keyfile_backup)

        went_fine = False
        try:
            _f = open(OCIConfigParser.LOCAL_KEY_FILE, 'w')
            _f.write(oci_keyfile)
            _f.close()
            logger.debug('new key file copied')
            went_fine = True
        except Exception, e:
            logger.warning('cannot save new key file [%s]' % str(e))
            raise StandardError('cannot save new key file [%s]' % str(e))
        finally:
            if os.path.exists(_keyfile_backup):
                if not went_fine:
                    shutil.copyfile(_keyfile_backup,
                                    OCIConfigParser.LOCAL_KEY_FILE)
                os.remove(_keyfile_backup)
        oci_fingerprint = compute_fingerprint(oci_keyfile_passphrase)
        if oci_fingerprint:
            config.set_option(OCIConfigParser.FINGERPRINT_OPT, oci_fingerprint)
        else:
            return RequestBadResult('failed to get fingerprint from keyfile').toDict()

    logger.debug('new user configuration %s' % config.to_safe_map())
    # first validate the config
    # config structure is the same as the one
    # described in OCI SDK configuration
    try:
        config.validate()
    except StandardError, e:
        return RequestBadResult('Invalid configuration: %s' % str(e)).toDict()

    logger.debug('new user configuration validated ')

    # now flush the config
    try:
        logger.debug('flushing new configuration')
        config.flush()
    except Exception, e:
        logger.warning('cannot save new configuration [%s]' % str(e))
        raise StandardError('cannot save new configuration [%s]' % str(e))

    return RequestGoodResult().toDict()


def delete_oci_config(params):
    """
    Deletes the OCI service access configuration for root user.
    Raise : StandardError if delete of configuration has failed
    """
    logger.debug('delete oci configuration')
    path = OCIConfigParser.LOCAL_CONF_HOME
    if os.path.exists(path):
        try:
            shutil.rmtree(path)
        except Exception as e:
            logger.exception('cannot delete oci config')
            return RequestBadResult('Cannot delete OCI configuration: %s' % str(e)).toDict()

    return RequestGoodResult().toDict()


def get_oci_config(params):
    """
    Returns the OCI service access configuration
    for root user. If configuration is missing we return an empty
    object (not treat as an error as configuration may not have been set yet)
    Returns a dict with configuration keys
    see Content/API/Concepts/sdkconfig.htm
    Note: security information are removed and not sent back
    Raise: StandardError if read of configuration has failed
    """
    if not os.path.exists(OCIConfigParser.CONFIG_FILE):
        logger.warning('conf file not found')
        return RequestGoodResult(dict()).toDict()
    config = OCIConfigParser()
    try:
        config.load()
    except Exception, e:
        logger.exception('cannot oci config')
        return RequestBadResult('Cannot read OCI configuration: %s' % str(e)).toDict()

    return RequestGoodResult(config.to_safe_map()).toDict()


def get_task(params):
    """
    Returns ongoing task list (can be empty)
    a task is
    {
       'ID': unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status': {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                  },
       'states': [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
    }
    """

    _id = params.get('ID')
    if not _id:
        return RequestBadResult('missing id param').toDict()

    try:
        return RequestGoodResult({'task': backend.taskmgr.get_task_manager().get_task(_id)}).toDict()
    except Exception, e:
        logger.exception('get_task for %s' % _id)
        return RequestBadResult('Cannot get task %d: %s' % (_id, str(e))).toDict()


def get_task_list(params):
    """
    Returns ongoing task list (can be empty)
    params:
       running_only: (optional) only running tasks ?
       'target_name': fetch only task for this target
       'task_name': fetch only task of that name
    a task is
    {
      'ID': unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status': {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                 },
       'states': [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
    }
    """
    running_only = bool(params.get('running_only', False))
    task_name_filter = params.get('task_name', None)

    task_target_name = params.get('target_name', None)

    logger.debug('get_task_list running_only=%s, task_name_filter=%s, task_target_name=%s'
                 % (running_only, task_name_filter, task_target_name,))

    if task_target_name is not None:
        all_tasks = backend.taskmgr.get_task_manager().get_task_for_target(task_target_name)
    else:
        all_tasks = backend.taskmgr.get_task_manager().get_tasks()
    r_tasks = []
    for t in all_tasks:
        if running_only:
            if not t.is_running():
                continue
        if task_name_filter is not None and t.get_name() != task_name_filter:
            continue
        if task_target_name is not None and t.get_target() != task_target_name:
            continue

        r_tasks.append(t)

    return RequestGoodResult({'tasks': r_tasks}).toDict()


def restart_nfs_stack(params):
    """
    Restart NFS stack
    raise:
        StandardError: restart (systemctl command) has failed
    """
    logger.debug('restart_nfs_stack')
    try:
        restart_nfs_services()
    except Exception as e:
        logger.exception('restart_nfs_stack failed')
        return RequestBadResult(e.args[0]).toDict()
    return RequestGoodResult().toDict()


def get_services_status(params):
    """
    get status of services
    expected argument:
        'service_names': comma separated list of service names
        'logs_required': if present include services'logs as part of the report
    returns:
        array of JSON object as follow
        {
         'status': [{
                      'name'   : <service name as string>
                      'dec'    : <service description as string>
                      'status' : <service status>
                      'logs'   : <service logs as string array>
                    },
                   ...]
        }
    see SystemdService.to_json()

    raise:
        StandardError: systemctl command has failed
    """

    serviceNames = params.get('service_names')
    logger.debug('get_services_status for [%s]' % serviceNames)
    if not serviceNames:
        return RequestBadResult('missing service_names param').toDict()
    logs_required = params.get('logs_required')

    allStatus = []
    for serviceName in [s.strip() for s in serviceNames.split(',') if s]:
        logger.debug('get status of [%s]' % serviceName)
        allStatus.append(json.dumps(SystemdService(serviceName).load_details((logs_required is not None)),
                                    cls=LSAExtJSONEncoder))

    return RequestGoodResult({'status': allStatus}).toDict()


def create_admin(params):
    ''' create the admin account with the password provided '''
    if 'password' not in params:
        return RequestBadResult('missing password param').toDict()

    (out, err, res) = execute_command([CMD_USER_ADD, 'admin', '-g', 'users'])
    if res != 0 and res != 12:
        return RequestBadResult('useradd command failed: %s' % err).toDict()
    (pout, perr, pres) = execute_command([CMD_PASSWD, '--stdin', 'admin'], input=params['password'])
    if pres != 0:
        (out, err, res) = execute_command([CMD_USER_DEL, '--remove', '--force', 'admin'])
        if res != 0:
            logger.error('cannot remove admin user [%s]' % err)
        return RequestBadResult('setting password for admin failed: %s' % perr).toDict()
    return RequestGoodResult().toDict()


LOG_SYSTEM_TYPE = 'system'
LOG_BOOT_TYPE = 'boot'
LOG_SERVICES_TYPE = 'services'
LOG_STORAGE_TYPE = 'storage'


def get_logs(params):
    '''
    returns the content of the log based on the log_type provided
    the params contains:
        log_type: the type of logs. can be 'system', 'boot', 'services', 'storage'
        keyword: regExp used to filter the content of logs
    '''
    logger.debug('get_logs called')

    log_type = params.get('log_type')
    if not log_type:
        return RequestBadResult('missing log_type parameter').toDict()

    keyword = params.get('keyword')
    content = ''

    cmdline = ['/bin/journalctl', '--no-pager', '--quiet', '--lines=' + str(DEFAULT_LOG_LINES), '-o', 'short']
    arg = ['--dmesg']  # default for LOG_STORAGE_TYPE and LOG_SYSTEM_TYPE

    if log_type == LOG_BOOT_TYPE:
        arg = ['--boot']
    if log_type == LOG_SERVICES_TYPE:
        arg = ['--unit=firewalld', '--unit=httpd', '--unit=nfs-server', '--unit=nfs-ipmapd', '--unit=nfs-mountd', '--unit=smb', '--unit=nmb']

    cmdline.extend(arg)

    try:
        content = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
        if keyword:
            content = do_filter(content, keyword)
    except (ValueError, OSError) as eos:
        logger.error("Failed to gather logs: %s" % str(eos))
        return RequestBadResult('Failed to gather logs: %s' % str(eos)).toDict()
    return RequestGoodResult({'content': content}).toDict()


def do_filter(content, keyword):
    '''
    filter the content of logs in terms of keyword
    return:
        filtered log
    parameter:
        content: logs not filtered yet
        keyword: regExp used to filter content
    '''
    output = StringIO.StringIO()
    lines = content.split('\n')
    for line in lines:
        if re.search(keyword, line, re.IGNORECASE):
            output.write(str(line) + '\n')
    filter_content = output.getvalue()
    output.close()
    return filter_content


def get_platform_version(params):
    """
    Get the platform version by `rpm -q --qf '%{VERSION}' linux-storage-manager`
    """
    logger.debug('get_platform_version called')
    cmdline = [CMD_RPM, '-q', '--qf', '%{VERSION}', 'linux-storage-manager']
    try:
        content = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
    except (ValueError, OSError) as eos:
        logger.error("Failed to get platform version: %s" % str(eos))
        return RequestBadResult('Failed to get platform version: %s' % str(eos)).toDict()
    return RequestGoodResult({'content': content}).toDict()


def get_disk_byid(device):
    """
    Get the /dev/disk/by-id entry of the device
    Input:
        device: name of the device, eg. sdb
    Output:
        the by-id entry of the device under /dev/disk/by-id.
        Note:
            1. If the device is a valid name, there must be an entry returned, otherwise, returns nothing(this would not happen).
            2. For each valid dev name, there actually are 2 entries mapping to it, and whichever returned is fine.
        eg. Here are the 2 entries under /dev/disk/by-id
            scsi-360b690ec43404b0f82423613a96ef5e7
            wwn-0x60b690ec43404b0f82423613a96ef5e7
        They both are symlink to the same dev ../../sdb. os.listdir() we use returns a list in arbitary order,
        so no guarantee which one of the two will be returned, but whichever one is good to describe the disk.
    """
    for devbyid in os.listdir(DEV_PATH_BY_ID):
        if os.path.basename(os.path.realpath(DEV_PATH_BY_ID + '/%s' % devbyid)) == device:
            return devbyid


def get_platform_disks(params):
    """
    Get the platform attached iscsi disks by calling oci_utils methods
    Return:
        a list of disks, each of which is described as follow
           name: name of the disk as string, eg. sda,sdb,etc
           id: /dev/disk/by-id entry of /dev/name
           type: type of the disk as string, ['iscsi' | 'fc' | 'sas' | 'nvmf' |...]
           wwn: world wide name as string
           size: size of disk in bytes as int
           state: state of disk as string, ['running' | 'not running']
           status: status of disk as dict
           {
               partitioned: status as boolean of whether the disk contains partitions
               mounted: status as boolean of whether the disk or at least one of the partitions(if any) is mounted
           }
    Raise:
        OSError or CalledProcessError when call to iscsiadm and/or lsblk failed
    """

    logger.debug('get_platform_disks called')
    disks = []
    attr = {}
    try:
        session = iscsiadm.session()

        # iscsiadm.session() returns a dict of targets attached, using
        # the IQNs as keys
        # { iqn1: { 'current_portal_ip': ip_address,
        #          'current_portal_port': port,
        #          'persistent_portal_ip': ip_address,
        #          'persistent_portal_port': port,
        #          'state': state,
        #          'device': sdX,
        #        },
        #  iqn2: {....}
        # }

    except (OSError, subprocess.CalledProcessError) as e:
        logger.error("Failed to execute /usr/sbin/iscsiadm: %s" % str(e))
        return RequestBadResult('Failed to execute /usr/sbin/iscsiadm: %s' % str(e)).toDict()
    try:
        blks = lsblk.list()

        # lsblk.list() returns a dict representing the scsi devices:
        # {device:
        #    {'mountpoint':mountpoint,
        #     'fstype':fstype,
        #     'size':size,
        #     'partitions':
        #           {device1:
        #               {'mountpoint':mountpoint1,
        #                'fstype':fstype1,
        #                'size':size1}
        #            device2:
        #               {'mountpoint':mountpoint2,
        #                'fstype':fstype2,
        #                'size':size2}
        #            ...
        #           }
        #    }
        # }

    except (OSError, subprocess.CalledProcessError) as e:
        logger.error("Failed to execute /usr/bin/lsblk: %s" % str(e))
        return RequestBadResult('Failed to execute /usr/bin/lsblk: %s' % str(e)).toDict()

    # get all physical volumes reachable on the platform
    _all_pvs = get_physical_volumes()
    logger.debug('all PVs found on the system : %s' % _all_pvs)
    try:
        for iqn in session.keys():
            device = session[iqn]['device']
            attr['name'] = device
            attr['id'] = get_disk_byid(device)
            attr['type'] = 'iscsi'
            attr['wwn'] = iqn
            attr['state'] = session[iqn]['state']
            attr['size'] = format_readable_to_bytes(blks[device]['size'])
            attr['status'] = {}
            attr['status']['is_pv'] = False
            logger.debug('/dev/%s' % device)
            if '/dev/%s' % device in _all_pvs:
                attr['status']['is_pv'] = True
            if 'partitions' in blks[device]:
                attr['status']['partitioned'] = True
                attr['status']['mounted'] = False
                for part in blks[device]['partitions'].values():
                    if part['mountpoint']:
                        attr['status']['mounted'] = True
                        break
            else:
                attr['status']['partitioned'] = False
                if blks[device]['mountpoint']:
                    attr['status']['mounted'] = True
                else:
                    attr['status']['mounted'] = False
            disks.append(attr)
            attr = {}
    except Exception as e:
        logger.error("key does not exist in oci-util returned dict: %s" % str(e))
        return RequestBadResult('key does not exist in oci-util returned dict: %s' % str(e)).toDict()
    logger.debug('%d disks returned' % len(disks))
    return RequestGoodResult({'disks': disks}).toDict()


def reboot(params):
    ''' reboot the system '''
    logger.debug('reboot called')
    res = subprocess.call([SYSCTL, 'reboot'])
    if res == 0:
        return RequestGoodResult().toDict()
    return RequestBadResult('systemctl command exited on %d' % res).toDict()


def poweroff(params):
    ''' power-off the system '''
    logger.debug('poweroff called')
    res = subprocess.call([SYSCTL, 'poweroff'])
    if res == 0:
        return RequestGoodResult().toDict()
    return RequestBadResult('systemctl command exited on %d' % res).toDict()


def update(params):
    ''' update appliance rpm and minimal security update (dont change versions)'''
    logger.debug('update called')

    cmdline = ['/usr/bin/yum', '-y', '--security', 'update-minimal']
    content = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
    logger.debug(content)
    if len(content) == 0:
        return RequestBadResult('failure running yum --security update-minimal').toDict()

    cmdline = ['/usr/bin/yum', '-y', 'install', 'update', 'linux-storage-manager.noarch']
    content_pkg_up = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
    logger.debug(content_pkg_up)
    if len(content_pkg_up) == 0:
        content_pkg_up = '\n failure running yum update linux-storage-manager.noarch'

    full_content = '=- Security update -=\n' + content + '\n=- Appliance update -=\n' + content_pkg_up
    logger.debug(full_content)
    return RequestGoodResult({'content': full_content}).toDict()


def get_ksplice_autoinstall_state(params):
    '''
    Get the current setting of ksplice auto install/update
    '''
    try:
        state = backend.system_util.get_ksplice_auto_enable_setting()
    except Exception, e:
        return RequestBadResult('could not retrieve ksplice state: %s' % str(e)).toDict()

    return RequestGoodResult({'autoinstall_state': state}).toDict()


def get_appliance_configuration_state(params):
    '''
    Check if the appliance is configured for storage usage.
    Returns: 'CONFIGURED'|'UNCONFIGURED'
    '''
    appliance_configured = 'UNCONFIGURED'
    try:
        if is_vg_created() is True:
            appliance_configured = 'CONFIGURED'
    except Exception, e:
        return RequestBadResult('Can not check if the volume group is created: %s' % str(e)).toDict()

    return RequestGoodResult({'state': appliance_configured}).toDict()


def toggle_ksplice_autoupdate_setting(params):
    '''
    Toggle the ksplice autoupdate settings (yes-->no, no-->yes)
    param:
        cur_sate: current setting of ksplice autoupdate
    raise: standard error in case of failure
    '''
    cur_state = params.get('cur_state')
    if not cur_state:
        return RequestBadResult('missing cur_state parameter').toDict()
    try:
        new_state = backend.system_util.toggle_ksplice_autoupdate_setting(cur_state)
    except Exception as e:
        return RequestBadResult('could not update ksplice settings: %s' % str(e)).toDict()
    return RequestGoodResult({'autoinstall_state': new_state}).toDict()


def get_vg_stats(params):
    '''
    Retrieve volume group usage details
    '''
    try:
        vg_stats = get_volume_group_stats()
    except Exception as e:
        logger.debug('failure in get_vg_stats: %s' % str(e))
        return RequestBadResult('could not retrieve volume group usage: %s' % str(e)).toDict()

    json_str = json.dumps(vg_stats)
    return RequestGoodResult({'vg_stats': json_str}).toDict()


def get_storage_pool(params):
    '''
    Get the storage pool info. At first we get the volume group lsa.
    If the volume group is backed by raid we add the device list from RadiArray otherwise we add the
    devices from volume group.
    Output: Serialized StoragePool, size(int), free(int), devices(list of RaidBacDev), protection(string), num_devs(int)
            format: {'size': '12345', 'free': '12345', 'status': 'active', 'devices': [{'index': '0', 'name': '/dev/sdb', state: 'active'}, {'index': '1', 'name': '/dev/sdc', state: 'active'}],
                      'protection': 'raidX', 'num_devs': 'X}
    '''
    try:
        vg_disks = list_vg_disks()
        vg_stats = get_volume_group_stats()
    except Exception as e:
        logger.exception('Exception on retrieving volume group information: %s' % str(e))
        return RequestBadResult('Failed to get volume group information: %s' % str(e)).toDict()
    try:
        if RAID_DEV in vg_disks:
            ra = backend.api_wrapper.StorageManager.get_storage_array()
            ra.set_vg_stats(vg_stats)
            sp = StoragePool.create_from_raid_array(ra)
        else:
            sp = StoragePool.create_from_vg(vg_stats, vg_disks)
        return RequestGoodResult({'pool': sp.serialize()}).toDict()
    except Exception as e:
        logger.debug('Failed to get storage pool: %s' % str(e))
        return RequestBadResult('Failed to get storage pool: %s' % str(e)).toDict()


def update_ssh_keys(params):
    '''
    Add ssh key to authorized_key file
    expected argument:
        'ssh_keys': public ssh key
        'user_key': user session key
    '''

    ssh_keys = params.get('ssh_keys')
    try:
        backend.ssh_util.update_ssh_keys(ssh_keys)
    except Exception as e:
        logger.debug('failure in adding ssh key: %s' % str(e))
        return RequestBadResult('Failed to add ssh key: %s' % str(e)).toDict()

    return RequestGoodResult().toDict()


def get_ssh_keys(params):
    try:
        ssh_keys = backend.ssh_util.get_ssh_keys()
        return RequestGoodResult({'ssh_keys': ssh_keys}).toDict()
    except Exception as e:
        logger.exception('failure to get ssh keys: %s' % str(e))
        return RequestBadResult('failure to get ssh keys: %s' % str(e)).toDict()


def set_samba_options(params):
    '''
    update samba global setting to samba conf file
    expected argument:
        'options': options -- workgroup, netbios name, interfaces
        'user_key': user session key
    '''

    options = params.get('options')
    try:
        smb_conf_manager = SambaConfigManager()
        smb_manager = SambaManager()
    except Exception, e:
        logger.exception('Failed to set samba options: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()

    try:
        smb_conf_manager.set_samba_options(options)
    except Exception as e:
        logger.debug('failure in setting samba options: %s' % str(e))
        return RequestBadResult('Failed to set samba options: %s' % str(e)).toDict()
    code = smb_manager.remove_export(None, None)
    if code != 0:
        smb_conf_manager.revert_smb_conf_modifications()
        smb_manager.restart_samba_services()
        return RequestBadResult('Failed to restart the services after global option modification, reverted the config file to previos persistent backup').toDict()

    return RequestGoodResult().toDict()


def get_samba_options(params):
    try:
        options = SambaConfigManager().get_samba_options()
        return RequestGoodResult({'options': options}).toDict()
    except Exception as e:
        logger.exception('failure to get samba options: %s' % str(e))
        return RequestBadResult('Failed to get samba options: %s' % str(e)).toDict()


def restart_samba_services(params):
    """
    Restart samba related services
    """
    try:
        smb_manager = SambaManager()
        smb_manager.restart_samba_services()
    except Exception, e:
        logger.exception('Failed to restart samba services: %s' % str(e))
        return RequestBadResult(e.args[0]).toDict()


def get_instance_details(params):
    '''
    get oci instance details
    Output JSON format:
        {'region': location of the instance(string)
         'domain': availability domain(string)
         'shape': type of shape(string)
         'instance_name': instance/display name(string)
         'creation_time': time when instance created(string)
        }
    '''
    inst_dict = {}
    try:
        metadata = oci_utils.metadata()
    except Exception as e:
        logger.exception('Failed to get oci metadata info: %s' % str(e))
        return RequestBadResult('Failed to get oci metadata info: %s' % str(e)).toDict()
    # check if the metadata is None, skip retrieving the items if None
    if metadata._metadata is not None:
        try:
            inst_dict['compartment_id'] = metadata['instance']['compartmentId']
            timeCreated = metadata['instance']['timeCreated']
            inst_dict['creation_time'] = time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime(timeCreated / 1000))
            inst_dict['region'] = oci_regions[metadata['instance']['region']]
            inst_dict['domain'] = metadata['instance']['availabilityDomain']
            inst_dict['shape'] = metadata['instance']['shape']
            inst_dict['instance_name'] = metadata['instance']['displayName']
        except Exception as e:
            logger.exception('Failed to get oci instance details: %s' % str(e))
            return RequestBadResult('Failed to get oci instance details: %s' % str(e)).toDict()

    return RequestGoodResult({'instance_details': inst_dict}).toDict()


def set_appliance_state(params):
    """
    set the appliance state
    expected option
      'state' :appliance state
    """
    state = params.get('state')
    try:
        subprocess.check_output([LSA_STATE_BIN, state],
                                stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError, e:
        logger.exception('Cannot set appliance state %s' % str(e))
        return RequestBadResult('Failed to set appliance state: %s' % str(e)).toDict()

    return RequestGoodResult().toDict()


def unconfigure_platform(params):
    '''
    Unconfigure the current platform.
    Following actions are taken:
       - stop/disable NFS server
       - stop/disable samba server
       - stop/disable http server
       - revert appliance state
    '''
    logger.debug('Stopping NFS server')
    SystemdService('nfs-server').load_details(False).disable_service(True)
    logger.debug('Stopping SMB server')
    SystemdService('smb').load_details(False).disable_service(True)
    SystemdService('nmb').load_details(False).disable_service(True)
    logger.debug('Stopping httpd server')
    SystemdService('httpd').load_details(False).disable_service(True)

    logger.debug('revert configuratin state file')
    set_appliance_state({'state': str(LSA_STATE_UNCONFIGURED)})
    return RequestGoodResult().toDict()


# command name to function mapping dictionary
# keys are
#   'fn': method to execute
#   'anonymous_allowed': can anonymous user execute it
cmd_mapping = {'authenticate': {'fn': authenticate, 'anonymous_allowed': True},
               'change_password': {'fn': change_password, 'anonymous_allowed': False},
               'get_storage_pool': {'fn': get_storage_pool, 'anonymous_allowed': False},
               'create_pool': {'fn': create_pool, 'anonymous_allowed': False},
               'create_share': {'fn': create_share, 'anonymous_allowed': False},
               'resize_share': {'fn': resize_share, 'anonymous_allowed': False},
               'delete_share': {'fn': delete_share, 'anonymous_allowed': False},
               'create_nfs_export': {'fn': create_nfs_export, 'anonymous_allowed': False},
               'modify_nfs_export': {'fn': modify_nfs_export, 'anonymous_allowed': False},
               'delete_nfs_export': {'fn': delete_nfs_export, 'anonymous_allowed': False},
               'create_samba_export': {'fn': create_samba_export, 'anonymous_allowed': False},
               'modify_samba_export': {'fn': modify_samba_export, 'anonymous_allowed': False},
               'delete_samba_export': {'fn': delete_samba_export, 'anonymous_allowed': False},
               'get_logs': {'fn': get_logs, 'anonymous_allowed': False},
               'get_platform_version': {'fn': get_platform_version, 'anonymous_allowed': False},
               'get_platform_disks': {'fn': get_platform_disks, 'anonymous_allowed': False},
               'get_instance_details': {'fn': get_instance_details, 'anonymous_allowed': False},
               'logout': {'fn': logout, 'anonymous_allowed': False},
               'test_priv': {'fn': test_priv, 'anonymous_allowed': False},
               'get_appliance_configuration_state': {'fn': get_appliance_configuration_state, 'anonymous_allowed': False},
               'list_shares': {'fn': list_shares, 'anonymous_allowed': False},
               'reboot': {'fn': reboot, 'anonymous_allowed': False},
               'update': {'fn': update, 'anonymous_allowed': False},
               'create_admin': {'fn': create_admin, 'anonymous_allowed': True},
               'admin_restart_nfs': {'fn': restart_nfs_stack, 'anonymous_allowed': False},
               'admin_restart_smb': {'fn': restart_samba_services, 'anonymous_allowed': False},
               'get_services_status': {'fn': get_services_status, 'anonymous_allowed': False},
               'get_vg_stats': {'fn': get_vg_stats, 'anonymous_allowed': False},
               'backup_share_dump': {'fn': backup_share_dump, 'anonymous_allowed': False},
               'restore_share_dump': {'fn': restore_share_dump, 'anonymous_allowed': False},
               'get_task_list': {'fn': get_task_list, 'anonymous_allowed': False},
               'get_task': {'fn': get_task, 'anonymous_allowed': False},
               'delete_oci_config': {'fn': delete_oci_config, 'anonymous_allowed': False},
               'get_oci_config': {'fn': get_oci_config, 'anonymous_allowed': False},
               'set_oci_config': {'fn': set_oci_config, 'anonymous_allowed': False},
               'update_ssh_keys': {'fn': update_ssh_keys, 'anonymous_allowed': False},
               'get_ssh_keys': {'fn': get_ssh_keys, 'anonymous_allowed': False},
               'set_samba_options': {'fn': set_samba_options, 'anonymous_allowed': False},
               'get_samba_options': {'fn': get_samba_options, 'anonymous_allowed': False},
               'get_ksplice_autoinstall_state': {'fn': get_ksplice_autoinstall_state, 'anonymous_allowed': False},
               'toggle_ksplice_autoupdate_setting': {'fn': toggle_ksplice_autoupdate_setting, 'anonymous_allowed': False},
               'unconfigure_platform': {'fn': unconfigure_platform, 'anonymous_allowed': False},
               'set_appliance_state': {'fn': set_appliance_state, 'anonymous_allowed': False}
               }


class RequestBadResult:

    def __init__(self, reason, code=0):
        """
        creates a new RequestBadResult
        reason: string message explaining the failure
        """
        self.cause = reason
        self.code = code

    def toDict(self):
        return {'results': 'failure', 'reason': self.cause, 'code': self.code}


class RequestGoodResult:
    def __init__(self, results=None):
        """
        creates a new RequestGoodResult
        return success
        parameter:
          optional dictionary to be passed as results
        """
        self.results = results

    def toDict(self):
        r = {'results': 'success'}
        if self.results:
            r.update(self.results)
        return r

    def __str__(self):
        return str(self.toDict())


def getChannel(sock):
    """
    server and client code is mixed within the
    same module but client use SafeSocket and server use
    socket.socket
    This helper will check for 'sock' type to retrieve the right
    channel
    """
    if getattr(sock, 'getSocket', None) is None:
        # server side type
        return sock
    # client side type
    return sock.getSocket()


def send_msg(sock, msg):
    # Prefix each message with a 4-byte length (network byte order)
    msg = struct.pack('>I', len(msg)) + msg
    getChannel(sock).sendall(msg)


def recv_msg(sock):
    # Read message length and unpack it into an integer
    raw_msglen = recv_all(sock, 4)
    if not raw_msglen:
        logger.warning('no message size received on socket')
        return None
    msglen = struct.unpack('>I', raw_msglen)[0]
    # Read the message data
    return recv_all(sock, msglen)


def recv_all(sock, n):
    # Helper function to receive n bytes or return None if EOF is hit
    data = ''
    while len(data) < n:
        packet = getChannel(sock).recv(n - len(data))
        if not packet:
            logger.warning('no data returned from socket aborting')
            logger.debug('data read so far [%s]' % data)
            return None
        data += packet
    return data


class Handler(asyncore.dispatcher_with_send):

    def handle_error(self):
        logger.exception('uncaptured exception in server')

    def handle_read(self):
        data = recv_msg(self)
        if data:
            # this can contain password: do not show this is production !
            # logger.debug('new data from socket [%s]' % data)
            cmd = json.loads(data)
            if 'cmd' not in cmd:
                send_msg(self, json.dumps({'results': 'failure', 'reason': 'missing command'}))
                return

            command_id = cmd['cmd']
            command_params = cmd['params']
            _cmd = cmd_mapping.get(command_id)
            if _cmd is None:
                send_msg(self, json.dumps({'results': 'failure', 'reason': 'unknown command [%s]' % command_id}))
                return

            if 'user_key' not in command_params or len(command_params['user_key']) == 0:
                send_msg(self, json.dumps({'results': 'failure', 'reason': 'missing or invalid user_key'}))
                return
            logger.debug('handle_read -> user_key=%s' % command_params['user_key'])
            logger.debug('handle_read -> anonymous allowed =%s' % _cmd['anonymous_allowed'])

            if _cmd['anonymous_allowed'] is False:
                if command_params['user_key'] == 'anonymous':
                    send_msg(self, json.dumps({'results': 'failure', 'reason': 'anonymous access not allowed'}))

            if command_params['user_key'] not in current_user_keys:
                send_msg(self, json.dumps({'results': 'failure', 'reason': 'invalid user_key'}))
                return

            res = _cmd['fn'](command_params)
            res['cmd'] = command_id
            send_msg(self, json.dumps(res, cls=LSAExtJSONEncoder))

        else:
            logger.warning('empty message returned from socket')


class Server(asyncore.dispatcher):

    def __init__(self, host, port):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(5)

    def handle_accept(self):
        logger.debug('waiting for new connection')
        pair = self.accept()
        if pair is not None:
            sock, addr = pair
            logger.info('Incoming connection from %s' % repr(addr))
            Handler(sock)


class SafeSocket():
    """
    wrapper class around socket.socket object
    To keep track of close state
    """
    def __init__(self, operationsTimeout):
        """
        Creates a new SafeSocket instance
        params:
          operationsTimeout: timeout set on the wrapped socket
        """
        self.opened = False
        self.opsTM = operationsTimeout
        self.socket = None

    def close(self):
        """
        change state to close and actually close the
        wrapped socket
        """
        self.socket.close()
        self.opened = False

    def connect(self, dest, port):
        """
        connect to remote host and change state to 'opened'
        params:
          dest: destination addr
          port: remote port to connect to
        """
        if self.opened:
            raise StandardError('already opened')

        self.socket = socket.socket()
        self.socket.settimeout(self.opsTM)
        self.socket.connect((dest, port))
        self.opened = True

    def getSocket(self):
        """
        Returns the wrapped socket
        """
        return self.socket

    def isOpened(self):
        """
        Returns the open state as boolean
        """
        return self.opened


# Blocking queue to store client socket
# for any request to the server, the queue is polled to retrieve
# the socket instance. the socket is put back to the queue at the end of the request
# which set the next waiting thread free to use the socket
clientQueue = Queue.Queue()

# for client side we don't have initialization phase. code is mixed with server side
# init the queue at import time
# On server side the queue will hold unused (unconnected) socket object
# default timeout SAFE_SOCKET_OP_TIMEOUT secs
clientQueue.put(SafeSocket(SAFE_SOCKET_OP_TIMEOUT))


def priv_exec(cmd, params, timeout=SAFE_SOCKET_OP_TIMEOUT):
    '''
    sends a message to the priv_exec_server
    params:
      cmd: the command to be sent as string
      params: command parameters as dict
      timeout: timeout waiting for communication channel to get available
                default value: SAFE_SOCKET_OP_TIMEOUT seconds
    returns:
      command result
    '''
    global clientQueue

    priv_client_socket = None
    try:
        priv_client_socket = clientQueue.get(True, timeout)
    except Queue.Empty:
        return RequestBadResult('timeout waiting for socket to be available').toDict()

    # logger.debug('acquired the channel')

    # socket may have been closed by previous call
    if not priv_client_socket.isOpened():
        logger.debug('channel not opened, connecting it')
        try:
            priv_client_socket.connect('127.0.0.1', 9991)
            logger.debug('connected to server')
        except socket.error, e:
            clientQueue.task_done()
            clientQueue.put(priv_client_socket)
            return RequestBadResult('socket connection failure [%s]' % str(e)).toDict()

    command = {'cmd': cmd, 'params': params}
    try:
        send_msg(priv_client_socket, json.dumps(command))
        # logger.debug('message sent waiting for inputs')

        reply = recv_msg(priv_client_socket)
        if not reply:
            # reply can be None in case of broken pipe
            result = RequestBadResult('socket error empty reply', 2).toDict()
        else:
            result = json.loads(reply)

    except Exception, e:
        logger.exception('socket error while receiving/sending message')
        # make sure next connection will be a new one
        priv_client_socket.close()
        result = RequestBadResult('socket error while sending message: %s' % str(e)).toDict()

    clientQueue.task_done()
    clientQueue.put(priv_client_socket)
    # logger.debug('command executed/channel released')
    return result


def main():
    try:
        # instantiate the task manager
        task_manager = TaskManager(os.path.join(VAR_DIR, TASK_MGR_DB_FILE))
        task_manager.load()
        backend.taskmgr.set_task_manager(task_manager)
        logger.debug('task manager started')
        Server('127.0.0.1', 9991)
        logger.debug('starting loop')
        asyncore.loop()
    except KeyboardInterrupt:
        pass
    else:
        logger.exception('exception raised in main loop')


PRIV_EXEC_LOG_BK_COUNT = 5
if __name__ == '__main__':
    try:
        set_log_config(PRIV_EXEC_LOG_FILENAME, PRIV_EXEC_LOG_BK_COUNT)
    except Exception, e:
        print 'cannot setup logging: %s' % str(e)

    logger.info('starting priv_exec_server')

    main()
