"""
module hosting setup blueprint views
all requests to https://<host>/setup arrive here
- storage setup
- admin setup
"""
__copyright__ = "Copyright 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from flask import Blueprint, render_template, request, redirect, url_for, flash, current_app
from flask.ext.login import login_required, current_user

from ..doc.views import auto
from ..security import create_admin_account, is_admin_created
from ..login.views import logout
import priv_exec_server as priv_exec
import backend_test as web_backend
from backend.common import RAID_LEVEL_NONE
from backend.common import LSA_STATE_READY

setup = Blueprint('setup', __name__,
                  template_folder='templates',
                  static_folder='static')


@setup.route('/storage', methods=['GET', 'POST'])
@login_required
def setup_storage_array():

    p_state = web_backend.get_appliance_configuration_state()
    if p_state == 'CONFIGURED':
        current_app.logger.warning('entering storage setup, while already in configured state')
        # return redirect(url_for('bootstrap.index'))

    all_disks = web_backend.get_platform_disks()

    all_usable_disks = []
    all_not_usable_disks = []
    # make two lists of disks: 'usable' and 'not usable'
    for d in all_disks:
        usable = True
        current_app.logger.debug('new disk %s' % d)
        if d['state'] != 'running':
            d['msg'] = 'Disk not in running state'
            current_app.logger.debug('discard not running disk %s [%s]' % (d['name'], d['state']))
            usable = False
        if d['status']['mounted'] is True:
            d['msg'] = 'Disk already mounted'
            current_app.logger.debug('discard mounted disk %s' % d['name'])
            usable = False
        if d['status']['is_pv'] is True:
            d['msg'] = 'Disk not clean, already initialized as a physical volume'
            current_app.logger.debug('discard PV disk %s' % d['name'])
            usable = False

        if usable is True:
            all_usable_disks.append(d)
        else:
            all_not_usable_disks.append(d)

    if request.method == 'GET':
        return render_template('storage.html',
                               disks=all_usable_disks,
                               not_available_disks=all_not_usable_disks)
    if request.method == 'POST':
        selected_device_list = request.form.getlist('dev_list_item')
        current_app.logger.debug('selected devices [%s]' % selected_device_list)
        try:
            web_backend.create_pool(RAID_LEVEL_NONE, ','.join(['/dev/%s' % d for d in selected_device_list]))
            # set appliance state to ready
            web_backend.set_appliance_state(LSA_STATE_READY)
            flash('Storage pool setup was successful, please login now')
            return redirect(url_for('bootstrap.index'))
        except Exception, e:
            current_app.logger.exception('failure %s' % str(e))
            return render_template('storage.html',
                                   error='Storage pool creation failed: %s' % str(e),
                                   disks=all_usable_disks,
                                   not_available_disks=all_not_usable_disks)


@setup.route('/admin', methods=['GET', 'POST'])
def create_admin():

    if is_admin_created():
        # admin already there , redirect directly
        current_app.logger.warning('admin already created, redirect to index')
        return redirect(url_for('bootstrap.index'))

    if request.method == 'GET':
        return render_template('create_admin.html')

    try:
        create_admin_account(request.form['new_password'])
        flash('Admin password setup was successful, please login now')
        return redirect(url_for('bootstrap.index'))
    except StandardError, e:
        return render_template('create_admin.html', creation_error='Error while setting admin password: %s' % str(e))


@setup.route("/changepassword", methods=['GET', 'POST'])
@auto.doc(groups=['internal'])
@login_required
def change_password():
    """
    change the password of admin account
    request parameters:
      username: username to change the password
      old_password: expected current user password
      new_password: new user password
    """
    current_app.logger.debug('inside /setup/changepassword for method [%s]' % request.method)

    if not is_admin_created():
        current_app.logger.debug('admin not already created, redirect to index')
        return redirect(url_for('bootstrap.index'))

    if request.method == 'GET':
        return render_template('password_change.html',
                               user_name=current_user.name)

    current_app.logger.debug('changing password')
    res = priv_exec.priv_exec('change_password',
                              {
                                  'username': request.form['username'],
                                  'old_password': request.form['old_password'],
                                  'new_password': request.form['new_password'],
                                  'user_key': current_user.user_key
                              })
    if res['results'] == 'success':
        """ password updated successfully, log current user out, and redirect to login page """
        current_app.logger.debug('password updated successfully, login out user')
        return redirect(logout())
    else:
        current_app.logger.debug('password updated failed')
        return render_template('password_change.html',
                               user_name=current_user.name,
                               creation_error=res['reason'])
